"""Functions for re-sampling Gaia data within errors."""

import numpy as np
import pandas as pd
from typing import Optional, Union


def generate_gaia_covariance_matrix(
    data_gaia: pd.DataFrame, six_parameter_sources: bool = False
) -> np.ndarray:
    """Generates a covariance matrix for Gaia data. Currently only has pmra/pmdec/parallax(/color) covariance matrix
    support, but could be easily extended in the future to also/or re-sample more things.

    Args:
        data_gaia (pd.DataFrame): Gaia data to make a covariance matrix for. Must contain the following keys:
                pmra_error, pmdec_error, parallax_error, pmra_pmdec_corr, parallax_pmra_corr, parallax_pmdec_corr
            For resampling for six parameter sources (i.e. six_parameter_sources=True), it must also contain:
                pseudocolour_error, pmra_pseudocolour_corr, pmdec_pseudocolour_corr, parallax_pseudocolour_corr
            see Gaia release notes for help.
        six_parameter_sources (bool): whether or not ALL sources in data_gaia also depend on the estimated pseudocolour.
            If true, will return matrices of shape (n_samples, 4, 4) instead.
            Default: False

    Returns:
        a covariance matrix of shape (n_samples, 3, 3) for pmra, pmdec and parallax (in that order), or (..., 4, 4) if
            six_parameter_sources is True.

    """

    # Grab the errors we need but as numpy arrays
    pmra = data_gaia["pmra_error"].to_numpy()
    pmdec = data_gaia["pmdec_error"].to_numpy()
    parallax = data_gaia["parallax_error"].to_numpy()

    # Also get the correlation coefficients, which come in the range [-1, 1]
    corr_pmra_pmdec = data_gaia["pmra_pmdec_corr"].to_numpy()
    corr_pmra_parallax = data_gaia["parallax_pmra_corr"].to_numpy()
    corr_pmdec_parallax = data_gaia["parallax_pmdec_corr"].to_numpy()

    # Calculate all of the terms
    pmra_pmra = pmra**2
    pmdec_pmdec = pmdec**2
    parallax_parallax = parallax**2

    pmra_pmdec = pmra * pmdec * corr_pmra_pmdec
    pmra_parallax = pmra * parallax * corr_pmra_parallax
    pmdec_parallax = pmdec * parallax * corr_pmdec_parallax

    # We have to do some extra steps if this source has a six-parameter solution (a DR3 thing).
    # Finally, we put it all into a big array (which we transpose from shape (3 or 4, 3 or 4, n_samples) to shape
    # (n_samples, 3 or 4, 3 or 4) to make later calculations more easily vectorisable)
    if six_parameter_sources:
        color = data_gaia["pseudocolour_error"].to_numpy()

        corr_pmra_color = data_gaia["pmra_pseudocolour_corr"].to_numpy()
        corr_pmdec_color = data_gaia["pmdec_pseudocolour_corr"].to_numpy()
        corr_parallax_color = data_gaia["parallax_pseudocolour_corr"].to_numpy()

        color_color = color**2

        pmra_color = pmra * color * corr_pmra_color
        pmdec_color = pmdec * color * corr_pmdec_color
        parallax_color = parallax * color * corr_parallax_color

        return np.asarray(
            [
                [pmra_pmra, pmra_pmdec, pmra_parallax, pmra_color],
                [pmra_pmdec, pmdec_pmdec, pmdec_parallax, pmdec_color],
                [pmra_parallax, pmdec_parallax, parallax_parallax, parallax_color],
                [pmra_color, pmdec_color, parallax_color, color_color],
            ]
        ).T
    
    return np.asarray(
        [
            [pmra_pmra, pmra_pmdec, pmra_parallax],
            [pmra_pmdec, pmdec_pmdec, pmdec_parallax],
            [pmra_parallax, pmdec_parallax, parallax_parallax],
        ]
    ).T


# Convenience vectorized definition of the multivariate_normal function. Unfortunately this isn't naturally vectorisable
# in numpy =(
_generator = np.random.default_rng()


def resample_gaia_astrometry(
    data_gaia: pd.DataFrame,
    check_valid: str = "ignore",
    method: str = "svd",
    n_resamples: int = 1,
    suffixes: Optional[Union[list, tuple, np.ndarray]] = None,
) -> pd.DataFrame:
    """Resample Gaia astrometric parameters for pmra, pmdec and parallax, given input best estimate means and covariance
    matrices. Numerous methods are available, with some faster than others!

    Todo: requires a unit test!

    Notes:
        - CHECKS ARE TURNED OFF TO MAKE THIS FUNCTION FASTER, so please make sure that data_gaia is correct and that
          no values have been messed up! No warranty is included on this BlazingFastTM function.
        - When using cholesky decomposition to speed this up, ALL COVARIANCE MATRICES MUST BE POSITIVE
          DEFINITE. Basically, this means they should be symmetric and have no negative numbers.

    Args:
        data_gaia (pd.DataFrame): data for the field, including keys 'pmra', 'pmdec' and 'parallax'.
            Shape (n_samples, :).
        check_valid (str, optional): whether or not to check if input matrices are valid for methods svd and eigh. May
            be one of 'warn’, ‘raise’ or ‘ignore’.
            Default: 'ignore'
        method (str, optional): method used by numpy.random.Generator.multivariate_normal to compute the factor. Can be
            one of:
            - 'svd': slowest, but recommended for stability at this time as some Gaia covariance matrices are quite bad
            - 'eigh': slightly faster
            - 'cholesky': fastest, but unstable (matrices MUST be positive definite - generally not applicable for Gaia)
            Default: 'svd'
        n_resamples (int): how many times to resample the astrometry, i.e. how many new values to return for each star.
            Default: 1
        suffixes (list-like, optional): suffixes to name each pmra, pmdec etc. with on return. If None, will name them
            _0, _1, ... etc.
            Default: None

    Returns:
        a new pd.DataFrame with the same indexing as data_gaia, with n_resamples new pmra, pmdec etc. columns.


    """
    # Firstly, let's identify which solutions are five or six parameter
    five_parameter_astrometry = data_gaia["astrometric_params_solved"].to_numpy() == 31
    six_parameter_astrometry = data_gaia["astrometric_params_solved"].to_numpy() == 95
    data_gaia_five = data_gaia.loc[five_parameter_astrometry]
    data_gaia_six = data_gaia.loc[six_parameter_astrometry]

    # Let's make somewhere to store solutions
    resampled_astrometry = np.empty((len(data_gaia), n_resamples, 3))

    # Generate the numpy vectorized instance (only a small speed increase but every little helps, also WHY THE FUCK IS
    # MULTIVARIATE NORMAL NOT ALREADY VECTORISABLE)
    # Todo: I think I know how to improve this to be vectorizable
    vectorized_multivariate_normal = np.vectorize(
        _generator.multivariate_normal,
        excluded={"size", "check_valid", "method"},
        cache=False,
        signature=f"(n),(n,n)->({n_resamples},n)",
    )

    # 5 parameter resampling - "straightforward"
    if np.any(five_parameter_astrometry):
        covariance_matrix_five = generate_gaia_covariance_matrix(data_gaia_five)

        resampled_astrometry[
            five_parameter_astrometry
        ] = vectorized_multivariate_normal(
            mean=data_gaia_five[["pmra", "pmdec", "parallax"]].to_numpy(),
            cov=covariance_matrix_five,
            size=n_resamples,
            check_valid=check_valid,
            method=method,
        )

    # 6 parameter resampling - we drop colours as the author of this script saw no need to use them =)
    if np.any(six_parameter_astrometry):
        covariance_matrix_six = generate_gaia_covariance_matrix(
            data_gaia_six, six_parameter_sources=True
        )

        resampled_astrometry[six_parameter_astrometry] = vectorized_multivariate_normal(
            mean=data_gaia_six[
                ["pmra", "pmdec", "parallax", "pseudocolour"]
            ].to_numpy(),
            cov=covariance_matrix_six,
            size=n_resamples,
            check_valid=check_valid,
            method=method,
        )[:, :, :-1]

    # Reshape the resampled astrometry into a dict, getting ready to turn it into a dataframe
    if suffixes is None:
        suffixes = [f"{x}" for x in range(n_resamples)]

    resampled_astrometry_dict = {}
    for i in range(n_resamples):
        resampled_astrometry_dict[f"pmra_{suffixes[i]}"] = resampled_astrometry[:, i, 0]
        resampled_astrometry_dict[f"pmdec_{suffixes[i]}"] = resampled_astrometry[
            :, i, 1
        ]
        resampled_astrometry_dict[f"parallax_{suffixes[i]}"] = resampled_astrometry[
            :, i, 2
        ]

    return pd.DataFrame(resampled_astrometry_dict)



"""Functions for doing various random sampling problems that aren't otherwise defined.
"""
import numpy as np
from typing import Tuple


def points_on_sphere(
    shape: int, radians: bool = True, phi_symmetric=True, seed=None
) -> Tuple[np.ndarray, np.ndarray]:
    """To draw random points on a sphere, one cannot simply use uniform deviates based on theta and phi! For a good
    explanation, see: http://mathworld.wolfram.com/SpherePointPicking.html

    Args:
        shape (int): required number of deviates. Should be acceptable by np.random.rand().
        radians (bool): whether or not to return in radians.
            Default: True
        phi_symmetric (bool): whether or not to return a distribution for phi symmetric about zero (so,
            phi ~ [-pi/2, pi/2]) or whether to return one defined in the mathsy (but less useful for astronomy) way
            phi ~ [0, pi].
            Default: True, i.e. phi is immediately usable as galactic latitude b or declination.

    Returns:
        two np.ndarray of random spherical deviates, for theta ~ [0, 2pi) and phi ~ [-pi/2, pi/2] (or in degrees if
            radians=False)

    """
    rng = np.random.default_rng(seed=seed)
    theta = 2 * np.pi * rng.uniform(size=shape)
    phi = np.arccos(2 * rng.uniform(size=shape) - 1)

    if phi_symmetric:
        phi -= np.pi / 2

    if radians:
        return theta, phi
    else:
        return theta * 180 / np.pi, phi * 180 / np.pi



"""The highest level within the plot sub-directory: a set of standard functions for making funky plots."""

from typing import Union, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .axis import cluster
from .axis import nn_statistics


def location(data_gaia: pd.DataFrame,
             show_figure: bool = True,
             save_name: Optional[str] = None,
             figure_size: Union[list, np.ndarray] = (10, 10),
             figure_title: str = None,
             dpi: int = 100,
             open_cluster_pm_to_mark: Optional[Union[list, np.ndarray]] = None,
             pmra_plot_limits: Optional[Union[list, np.ndarray]] = None,
             pmdec_plot_limits: Optional[Union[list, np.ndarray]] = None,
             plot_std_limit: float = 1.5,
             kde_resolution: int = 50,
             plotting_resolution: int = 50,
             kde_bandwidth_radec: float = 0.08,
             kde_bandwidth_pmotion: float = 0.08):
    """A figure showing the location of an open cluster in position and proper motion space. Has more parameters than
    there are grains of sand on all the beaches on the planet.

    Args:
        --- Function unique ---
        data_gaia (pandas.DataFrame): the Gaia data read in to a DataFrame.
            Keys should be unchanged from default Gaia source table names.
        open_cluster_pm_to_mark (list-like, optional): the co-ordinates
            (pmra, pmdec) of a point to mark on the proper motion diagram, such
            as a literature value.
        pmra_plot_limits (list-like, optional): the minimum and maximum
            proper motion in the right ascension direction plot limits.
        pmdec_plot_limits (list-like, optional): the minimum and maximum
            proper motion in the declination direction plot limits.
        plot_std_limit (float): standard deviation of proper motion to use
            to find plotting limits if none are explicitly specified.
            Default: 1.5
        kde_resolution (int): number of points to sample the KDE at when scoring
            (across a resolution x resolution-sized grid.)
            Default: 50.
        kde_bandwidth_radec (float): the bandwidth of the position kde.
            Default: 0.08
        kde_bandwidth_pmotion (float): the bandwidth of the proper motion kde.
            Default: 0.08
        plotting_resolution (int): number of levels to use in contour plots.
            Default: 50

        --- Module standard ---
        show_figure (bool): whether or not to show the figure at the end of plotting. Default: True
        save_name (string, optional): whether or not to save the figure at the end of plotting.
            Default: None (no figure is saved)
        figure_size (list-like): the 2D size of the figure in inches, ala Matplotlib. Default: (10, 10).
        figure_title (string, optional): the desired title of the figure.
        dpi (integer): dpi for display and saving. 300 is nice for high quality figures or when you need to see detail.
            Default: 100

    Returns:
        The generated figure.

    """
    # Initialise the figure and add stuff to the axes
    fig, ax = plt.subplots(nrows=2, ncols=2, figsize=figure_size, dpi=dpi)

    ax[0, 0], ax[0, 1] = cluster.position_and_pmotion(
        fig,
        ax[0, 0],
        ax[0, 1],
        data_gaia,
        open_cluster_pm_to_mark=open_cluster_pm_to_mark,
        pmra_plot_limits=pmra_plot_limits,
        pmdec_plot_limits=pmdec_plot_limits,
        plot_std_limit=plot_std_limit)

    ax[1, 0], ax[1, 1] = cluster.density_position_and_pmotion(
        ax[1, 0],
        ax[1, 1],
        data_gaia,
        kde_bandwidth_radec=kde_bandwidth_radec,
        kde_bandwidth_pmotion=kde_bandwidth_pmotion,
        pmra_plot_limits=pmra_plot_limits,
        pmdec_plot_limits=pmdec_plot_limits,
        plot_std_limit=plot_std_limit,
        kde_resolution=kde_resolution,
        plotting_resolution=plotting_resolution)

    # Beautifying
    fig.subplots_adjust(hspace=0.25, wspace=0.25)

    # Plot a title - we use text instead of the title UI so that it can be long and multi-line.
    if figure_title is not None:
        ax[0, 0].text(0., 1.10, figure_title, transform=ax[0, 0].transAxes, va="bottom")

    # Output time
    if save_name is not None:
        fig.savefig(save_name, dpi=dpi)

    if show_figure is True:
        fig.show()

    return fig, ax


_default_position_kwargs = {
    'x_limits': 'all_stars',  # one of: ['like_y', 'all_stars', 'all_clusters', [min, max], float_for_std_deviation_about_mean]
    'y_limits': 'x',  # one of: ['like_x', 'all_stars', 'all_clusters', [min, max], float_for_std_deviation_about_mean]
    'x_label': 'ra',  # String
    'y_label': 'dec',  # String
    'title': 'position',  # String
    'alpha': 'infer',  # Either 'infer' or a float between 0 and 1
    'legend': False,  # Whether or not to have a legend
    'legend_kwargs': {},  # Should be acceptable args of plt.legend()
    'extra_objects_to_plot': {},  # Dicts of format (n_samples, 2) to plot, where the key will be set to the label
    'colour_map': None,  # Which colour cycle to use when plotting clusters. None uses the default mpl style.
    'key_x': 'ra',  # Key into data_gaia
    'key_y': 'dec',  # Key into data_gaia. Ignored if type=histogram
    'marker_radius': 1.,  # Float
    'background_type': 'scatter',  # Type of plot. Allows 'scatter', 'line', 'kde', 'histogram', 'histogram_2d', or None to ignore.
    'cluster_type': 'scatter',  # As above
}


def _get_figure_shape(format: Union[tuple, list, np.ndarray]):
    """Uses a format list to get and check the shape of a figure."""
    # Cast to a numpy array


    pass


def clustering_result_new(data_gaia: pd.DataFrame,
                      cluster_labels: Optional[np.ndarray] = None,
                      cluster_indices: Optional[Union[list, np.ndarray]] = None,
                      cluster_shading: Optional[np.ndarray] = None,

                      format: Union[tuple, list, np.ndarray] = (('position', 'pmotion',),
                                                                ('cmd', 'blank')),

                      axis_kwargs: Optional[Union[list, tuple, np.ndarray]] = None,

                      show_figure: bool = True,
                      save_name: Optional[str] = None,
                      figure_size: Union[list, np.ndarray, tuple] = (10, 10),
                      figure_title: str = None,
                      dpi: int = 100,
                      ):
    """A figure for evaluating the results of a clustering algorithm.

    todo: could spec something that calls add_subplot() instead of using the format kwarg, making this much more
        flexible as a format for other plotting functions too

    Parameters
    ----------
    data_gaia : pd.DataFrame, shape (n_field_stars, n_features)
        the Gaia data read in to a DataFrame. Keys should be unchanged from default Gaia source table names.
    cluster_labels : np.ndarray, optional, shape (n_field_stars,)
        the cluster membership labels for clustered stars. This should be the default sklearn.cluster output (i.e.
        -1 is used to label field stars).
    cluster_indices : np.ndarray, optional, shape (n_clusters_to_plot,)
        the clusters to plot in cluster_labels. For instance, you may not want to plot '-1'.
    cluster_shading : np.ndarray, optional, shape (n_field_stars,)
        an array of floats in the range 0, 1 for all clusters to use to shade their colour (aka alpha value). Useful for
        displaying e.g. HDBSCAN soft clustering.
        Default: None
    format : list-like of str
        a 2D array where each entry contains the type of plot to make on each axis. The number of axes is inferred from
        the shape of the array. Current allowed axis types:
            'position'
            'proper_motion'
            'cmd' (colour magnitude diagram)
            'parallax_density' (a histogram of parallax values)
            'blank' (plots nothing)
        Default: (('position', 'pmotion',), ('cmd', 'parallax_density')), i.e. a position, pmotion, cmd, parallax_density plot.
    axis_kwargs : list-like of dicts, optional
        any parameters to overwrite for each axis. If specified, it should have the same shape as format.
        Default: None

    show_figure : bool
        whether or not to show the figure at the end of plotting.
        Default: True
    save_name : str, optional
        if specified, the name to save the file as. The file extension present (or .png) will be used.
        Default: None
    figure_size : list-like, shape (2,)
        the size of the figure (x, y) in inches.
        Default: 10, 10
    figure_title : str, optional
        supertitle for the entire figure object.
        Default: None
    dpi : int
        dots per inch for the figure object and for saving.
        Default: None

    Raises
    ------

    Returns
    -------
    fig : the matplotlib figure element
    ax : the matplotlib list of axes, with the same shape as format

    """
    # Initialise the figure
    fig, ax = plt.subplots(nrows=2, ncols=2, figsize=figure_size, dpi=dpi)



    pass


def clustering_result(data_gaia: pd.DataFrame,
                      cluster_labels: Optional[np.ndarray] = None,
                      cluster_indices: Optional[Union[list, np.ndarray]] = None,
                      cluster_shading: Optional[np.ndarray] = None,
                      show_figure: bool = True,
                      save_name: Optional[str] = None,
                      figure_size: Union[list, np.ndarray] = (10, 10),
                      figure_title: str = None,
                      dpi: int = 100,
                      open_cluster_pm_to_mark: Optional[Union[list, np.ndarray]] = None,
                      pmra_plot_limits: Optional[Union[list, np.ndarray]] = None,
                      pmdec_plot_limits: Optional[Union[list, np.ndarray]] = None,
                      cmd_plot_x_limits: Optional[Union[list, np.ndarray]] = None,
                      cmd_plot_y_limits: Optional[Union[list, np.ndarray]] = None,
                      cmd_plot_color_key: bool = None,
                      parallax_plot_x_limits: Optional[Union[list, np.ndarray]] = None,
                      parallax_plot_y_limits: Optional[Union[list, np.ndarray]] = None,
                      plot_std_limit: float = 1.5,
                      cmd_plot_std_limit: float = 3.0,
                      cluster_marker_radius: Union[tuple, list] = (1., 1., 1., 1.),
                      clip_to_fit_clusters: bool = True,
                      make_parallax_plot: bool = False):
    """A figure for evaluating the results of a clustering algorithm.

    Args:
        data_gaia (pandas.DataFrame): the Gaia data read in to a DataFrame.
            Keys should be unchanged from default Gaia source table names.
        cluster_labels (np.ndarray): the cluster membership labels for
            clustered stars. This should be the default sklearn.cluster output.
        cluster_indices (list-like, optional): the clusters to plot in
            cluster_labels. For instance, you may not want to plot '-1'
        cluster_shading (np.ndarray, optional): an array of floats in the range 0, 1 for all clusters to use to shade
            their colour (aka alpha value). Useful for displaying e.g. HDBSCAN soft clustering.
            Default: None
        open_cluster_pm_to_mark (list-like, optional): the co-ordinates
            (pmra, pmdec) of a point to mark on the proper motion diagram, such
            as a literature value.
        pmra_plot_limits (list-like, optional): the minimum and maximum
            proper motion in the right ascension direction plot limits.
        pmdec_plot_limits (list-like, optional): the minimum and maximum
            proper motion in the declination direction plot limits.
        cmd_plot_x_limits (list-like, optional): the minimum and maximum x (color) limits in the cmd plot.
        cmd_plot_y_limits (list-like, optional): the minimum and maximum y (apparent magnitude) limits in the cmd plot.
        cmd_plot_color_key (string, optional): optional other color in data_gaia to use instead of calculating G - RP.
        parallax_plot_x_limits (list-like, optional): as above for ra on the parallax plot.
        parallax_plot_y_limits (list-like, optional): as above for parallax on the parallax plot.
        plot_std_limit (float): standard deviation of proper motion to use to find plotting limits if none are
            explicitly specified.
            Default: 1.5
        cmd_plot_std_limit (float): standard deviation of proper motion to use to find plotting limits if none are
            explicitly specified, for the colour magnitude diagram plot. This is a separate parameter, as a higher
            value is often more suitable here.
            Default: 3.0
        cluster_marker_radius (list-like): radius of the cluster markers. Useful to increase when clusters are hard to
            see against background points. Specified as a length 4 tuple, giving the radius for the position, pm, CMD
            and parallax plots.
            Default: (1., 1., 1.)
        clip_to_fit_clusters (bool): whether or not to ensure that the proper motion plot always includes all clusters.
            Default: True
        make_parallax_plot (bool): whether to make a parallax plot on the fourth axes or whether to leave them blank.
            Default: False (for backwards-compatibility)

        --- Module standard ---
        show_figure (bool): whether or not to show the figure at the end of plotting. Default: True
        save_name (string, optional): whether or not to save the figure at the end of plotting.
            Default: None (no figure is saved)
        figure_size (list-like): the 2D size of the figure in inches, ala Matplotlib. Default: (10, 10).
        figure_title (string, optional): the desired title of the figure.
        dpi (integer): dpi for display and saving. 300 is nice for high quality figures or when you need to see detail.
            Default: 100

    Returns:
        The generated figure.

    """
    # Initialise the figure and add stuff to the axes
    fig, ax = plt.subplots(nrows=2, ncols=2, figsize=figure_size, dpi=dpi)

    ax[0, 0], ax[0, 1] = cluster.position_and_pmotion(
        fig, ax[0, 0], ax[0, 1], data_gaia, cluster_labels, cluster_indices, cluster_shading,
        open_cluster_pm_to_mark=open_cluster_pm_to_mark,
        pmra_plot_limits=pmra_plot_limits,
        pmdec_plot_limits=pmdec_plot_limits,
        plot_std_limit=plot_std_limit,
        cluster_marker_radius=cluster_marker_radius[0:2],
        clip_to_fit_clusters=clip_to_fit_clusters)

    ax[1, 0] = cluster.color_magnitude_diagram(
        fig, ax[1, 0], data_gaia, cluster_labels, cluster_indices, cluster_shading,
        x_limits=cmd_plot_x_limits,
        y_limits=cmd_plot_y_limits,
        plot_std_limit=cmd_plot_std_limit,
        cluster_marker_radius=cluster_marker_radius[2],
        cmd_plot_color_key=cmd_plot_color_key
    )

    if make_parallax_plot:
        ax[1, 1] = cluster.ra_versus_parallax(
            fig, ax[1, 1], data_gaia, cluster_labels, cluster_indices, cluster_shading,
            x_limits=parallax_plot_x_limits,
            y_limits=parallax_plot_y_limits,
            plot_std_limit=plot_std_limit,
            cluster_marker_radius=cluster_marker_radius[3]
        )

    # Beautifying
    fig.subplots_adjust(hspace=0.25, wspace=0.25)

    # Plot a title - we use text instead of the title api so that it can be long and multi-line.
    if figure_title is not None:
        ax[0, 0].text(0., 1.10, figure_title, transform=ax[0, 0].transAxes, va="bottom")

    # Output time
    if save_name is not None:
        fig.savefig(save_name, dpi=dpi)

    if show_figure is True:
        fig.show()

    return fig, ax


def nearest_neighbor_distances(distances: np.ndarray,
                               number_of_derivatives: int = 0,
                               show_figure: bool = True,
                               save_name: Optional[str] = None,
                               figure_size: Union[list, np.ndarray] = (6, 6),
                               figure_title: str = None,
                               dpi: int = 100,
                               **kwargs):
    """A function for plotting statistics about the kth nearest neighbor distance. Wraps the function
    ocelot.plot.axis.nn_statistics.point_number_vs_nn_distance.

    Args:
        --- Function unique: for plotting on the axes ---
        distances (np.ndarray): the kth nearest neighbor distance array, of shape (n_samples, n_neighbors_calculated.))
        number_of_derivatives (int): the number of derivatives to plot.
            Default: 0 (just plots the nn distances themselves)
        **kwargs: additional keyword arguments to pass to ocelot.plot.axis.nn_statistics.point_number_vs_nn_distance

        --- Module standard ---
        show_figure (bool): whether or not to show the figure at the end of plotting.
            Default: True
        save_name (string, optional): whether or not to save the figure at the end of plotting.
            Default: None (no figure is saved)
        figure_size (list-like): the 2D size of the figure in inches, ala Matplotlib.
            Default: (6, 6).
        figure_title (string, optional): the desired title of the figure.
        dpi (integer): dpi for display and saving. 300 is nice for high quality figures or when you need to see detail.
            Default: 100

    Returns:
        a list of the matplotlib figure, the matplotlib axis

    """

    fig, ax = plt.subplots(nrows=number_of_derivatives + 1, ncols=1, figsize=figure_size, sharex='all', dpi=dpi)

    # Do the plotty thing
    ax = nn_statistics.point_number_vs_nn_distance(ax, distances, **kwargs)

    # Plot a title - we use text instead of the title api so that it can be long and multi-line.
    if figure_title is not None:
        if type(ax) is np.ndarray:
            ax[0].text(0., 1.10, figure_title, transform=ax[0].transAxes, va="bottom")
        else:
            ax.text(0., 1.10, figure_title, transform=ax.transAxes, va="bottom")

    # Output time
    if save_name is not None:
        fig.savefig(save_name, dpi=dpi)

    if show_figure is True:
        fig.show()

    return fig, ax



"""A set of functions for adding standardised things to an axis, specifically for star cluster plotting."""

from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .. import process
from .. import utilities


def position_and_pmotion(
        fig,
        axis_1,
        axis_2,
        data_gaia: pd.DataFrame,
        cluster_labels: np.ndarray = None,
        cluster_indices: Optional[Union[list, np.ndarray]] = None,
        cluster_shading: Optional[np.ndarray] = None,
        open_cluster_pm_to_mark: Optional[Union[list, np.ndarray]] = None,
        pmra_plot_limits: Optional[Union[list, np.ndarray]] = None,
        pmdec_plot_limits: Optional[Union[list, np.ndarray]] = None,
        plot_std_limit: float = 1.5,
        cluster_marker_radius: Union[list, tuple] = (1., 1.),
        clip_to_fit_clusters: bool = True):
    """Makes a scatter plot of position and proper motion for a given cluster,
    using plot_helper_calculate_alpha to prevent over-saturation of the
    figures.

    Args:
        fig (matplotlib figure): the figure element. Required to calculate alpha values precisely.
        axis_1 (matplotlib axis): the positional (ra/dec) axis.
        axis_2 (matplotlib axis): the proper motion (pmra/pmdec) axis.
        data_gaia (pandas.DataFrame): the Gaia data read in to a DataFrame.
            Keys should be unchanged from default Gaia source table names.
        cluster_labels (np.ndarray): the cluster membership labels for
            clustered stars. This should be the default sklearn.cluster output.
        cluster_indices (list-like, optional): the clusters to plot in
            cluster_labels. For instance, you may not want to plot '-1'
            clusters (which are noise) produced by algorithm like DBSCAN.
        cluster_shading (np.ndarray, optional): an array of floats in the range 0, 1 for all clusters to use to shade
            their colour (aka alpha value). Useful for displaying e.g. HDBSCAN soft clustering.
            Default: None
        open_cluster_pm_to_mark (list-like, optional): the co-ordinates (pmra, pmdec) of a point to mark on the proper
            motion diagram, such as a literature value.
        pmra_plot_limits (list-like, optional): the minimum and maximum proper motion in the right ascension direction
            plot limits.
        pmdec_plot_limits (list-like, optional): the minimum and maximum proper motion in the declination direction
            plot limits.
        plot_std_limit (float): standard deviation of proper motion to use to find plotting limits if none are
            explicitly specified. Default: 1.5
        cluster_marker_radius (float): radius of the cluster markers. Useful to increase when clusters are hard to
            see against background points. Specified as a length 2 tuple, giving the radius for the position and pm
            plots.
            Default: (1., 1.)

    Returns:
        axis_1 (matplotlib axis): the modified position axis.
        axis_2 (matplotlib axis): the modified proper motion axis.
    """
    # Calculate our own plotting limits if none have been specified
    if pmra_plot_limits is None:
        pmra_plot_limits = (np.mean(data_gaia.pmra)
                            + np.std(data_gaia.pmra)
                            * np.array([-plot_std_limit, plot_std_limit]))

    if pmdec_plot_limits is None:
        pmdec_plot_limits = (np.mean(data_gaia.pmdec)
                             + np.std(data_gaia.pmdec)
                             * np.array([-plot_std_limit, plot_std_limit]))

    # We also clip the plot so that it will always include all found clusters
    if cluster_labels is not None and clip_to_fit_clusters:
        clustered_stars = np.isin(cluster_labels, cluster_indices)
        pmra_plot_limits = np.asarray(
            [np.clip(pmra_plot_limits[0], -np.inf, data_gaia.loc[clustered_stars, 'pmra'].min() - 1),
             np.clip(pmra_plot_limits[1], data_gaia.loc[clustered_stars, 'pmra'].max() + 1, np.inf)])
        pmdec_plot_limits = np.asarray(
            [np.clip(pmdec_plot_limits[0], -np.inf, data_gaia.loc[clustered_stars, 'pmdec'].min() - 1),
             np.clip(pmdec_plot_limits[1], data_gaia.loc[clustered_stars, 'pmdec'].max() + 1, np.inf)])

    if cluster_shading is None:
        cluster_shading = np.ones(data_gaia.shape[0])

    # Better notation
    pmra_range = pmra_plot_limits
    pmdec_range = pmdec_plot_limits

    # Work out how many points will fall in the pm range
    good_pmra = np.logical_and(data_gaia['pmra'] > pmra_range[0], data_gaia['pmra'] < pmra_range[1])
    good_pmdec = np.logical_and(data_gaia['pmdec'] > pmdec_range[0], data_gaia['pmdec'] < pmdec_range[1])
    pm_plot_points = np.count_nonzero(np.logical_and(good_pmra, good_pmdec))

    # Calculate a rough guess at a good alpha value
    alpha_estimate_1 = utilities.calculate_alpha(fig, axis_1, data_gaia.shape[0], 1)
    alpha_estimate_2 = utilities.calculate_alpha(fig, axis_1, pm_plot_points, 1)

    # Ra/dec plot
    axis_1.plot(data_gaia['ra'], data_gaia['dec'], '.', ms=1, alpha=alpha_estimate_1, c='k')
    axis_1.set_xlabel('ra')
    axis_1.set_ylabel('dec')
    axis_1.set_title('position')

    # Proper motion plot
    axis_2.plot(data_gaia['pmra'], data_gaia['pmdec'], '.', ms=1, alpha=alpha_estimate_2, c='k')
    axis_2.set_xlabel('pmra')
    axis_2.set_ylabel('pmdec')
    axis_2.set_xlim(pmra_range)
    axis_2.set_ylim(pmdec_range)
    axis_2.set_title('proper motion')

    # Plot clusters if labels have been specified
    if cluster_labels is not None:

        cmap = plt.get_cmap("tab10")  # Get the default matplotlib colourmap

        # We cycle over the clusters, grabbing their indices and plotting
        # them in a new colour erri time =)
        for i_color, a_cluster in enumerate(cluster_indices):
            stars_in_this_cluster = cluster_labels == a_cluster

            # Make colors and give them alpha values
            colors = np.tile(cmap(i_color % 10), (np.count_nonzero(stars_in_this_cluster), 1))
            colors[:, 3] = cluster_shading[stars_in_this_cluster]

            axis_1.scatter(data_gaia.loc[stars_in_this_cluster, 'ra'],
                           data_gaia.loc[stars_in_this_cluster, 'dec'],
                           marker='.',
                           s=cluster_marker_radius[0]**2,
                           c=colors,
                           zorder=100,
                           label=a_cluster)

            axis_2.scatter(data_gaia.loc[stars_in_this_cluster, 'pmra'],
                           data_gaia.loc[stars_in_this_cluster, 'pmdec'],
                           marker='.',
                           s=cluster_marker_radius[1]**2,
                           c=colors,
                           zorder=100,
                           label=a_cluster)

    # Add a red cross at a defined location on the pmra/pmdec plot if desired
    if open_cluster_pm_to_mark is not None:
        axis_2[0, 1].plot(open_cluster_pm_to_mark[0], open_cluster_pm_to_mark[1],
                          'rx', ms=6)

    return [axis_1, axis_2]


def density_position_and_pmotion(axis_1,
                                 axis_2,
                                 data_gaia,
                                 kde_bandwidth_radec,
                                 kde_bandwidth_pmotion,
                                 pmra_plot_limits=None,
                                 pmdec_plot_limits=None,
                                 plot_std_limit=1.5,
                                 kde_resolution: int = 50,
                                 plotting_resolution: int = 50):
    """Makes KDE plots of position density and proper motion density for a
    given cluster.

    Args:
        axis_1 (matplotlib axis): the positional (ra/dec) axis.
        axis_2 (matplotlib axis): the proper motion (pmra/pmdec) axis.
        data_gaia (pandas.DataFrame): the Gaia data read in to a DataFrame.
            Keys should be unchanged from default Gaia source table names.
        kde_bandwidth_radec (float): the bandwidth of the kde for position.
        kde_bandwidth_pmotion (float): the bandwidth of the proper motion
            kde.
        pmra_plot_limits (list-like, optional): the minimum and maximum
            proper motion in the right ascension direction plot limits.
        pmdec_plot_limits (list-like, optional): the minimum and maximum
            proper motion in the declination direction plot limits.
        plot_std_limit (float): standard deviation of proper motion to use
            to find plotting limits if none are explicitly specified.
            Default: 1.5
        kde_resolution (int): number of points to sample the KDE at when scoring
            (across a resolution x resolution-sized grid.) Default: 50.
        plotting_resolution (int): number of levels to use in contour plots.
            Default: 50

    Returns:
        axis_1 (matplotlib axis): the modified position axis.
        axis_2 (matplotlib axis): the modified proper motion axis.

    """
    # Calculate our own plotting limits if none have been specified
    if pmra_plot_limits is None:
        pmra_plot_limits = (np.mean(data_gaia['pmra'])
                            + np.std(data_gaia['pmra'])
                            * np.array([-plot_std_limit, plot_std_limit]))

    if pmdec_plot_limits is None:
        pmdec_plot_limits = (np.mean(data_gaia['pmdec'])
                             + np.std(data_gaia['pmdec'])
                             * np.array([-plot_std_limit, plot_std_limit]))

    # Better notation
    pmra_range = pmra_plot_limits
    pmdec_range = pmdec_plot_limits

    # Process the datasets - we may want to cut down the pmra/pmdec one
    data_position = data_gaia[['ra', 'dec']].to_numpy()

    data_proper_motion = data_gaia[['pmra', 'pmdec']]

    # Drop any bad proper motion points
    good_pmra = np.logical_and(data_gaia['pmra'] > pmra_range[0], data_gaia['pmra'] < pmra_range[1])
    good_pmdec = np.logical_and(data_gaia['pmdec'] > pmdec_range[0], data_gaia['pmdec'] < pmdec_range[1])
    good_stars = np.logical_and(good_pmra, good_pmdec)
    data_proper_motion = data_proper_motion.loc[good_stars, :].to_numpy()

    # Perform KDE fits to the data
    mesh_ra, mesh_dec, density_ra_dec = process.kde_fit_2d(data_position, kde_bandwidth_radec,
                                                           kde_resolution=kde_resolution)

    mesh_pmra, mesh_pmdec, density_pmotion = process.kde_fit_2d(data_proper_motion, kde_bandwidth_pmotion,
                                                                x_limits=pmra_range, y_limits=pmdec_range,
                                                                kde_resolution=kde_resolution)

    # Plot the results
    # Ra/dec
    axis_1.contourf(mesh_ra, mesh_dec,
                    density_ra_dec,
                    levels=plotting_resolution,
                    cmap=plt.get_cmap('Reds'))
    axis_1.set_xlabel('ra')
    axis_1.set_ylabel('dec')
    axis_1.text(0.02, 1.02, f'bw={kde_bandwidth_radec:.4f}',
                ha='left', va='bottom', transform=axis_1.transAxes,
                fontsize='small',
                bbox=dict(boxstyle='round', facecolor='w', alpha=0.8))

    # Pmra/pmdec
    axis_2.contourf(mesh_pmra, mesh_pmdec,
                    density_pmotion,
                    levels=plotting_resolution,
                    cmap=plt.get_cmap('Blues'))
    axis_2.set_xlabel('pmra')
    axis_2.set_ylabel('pmdec')
    axis_2.text(0.02, 1.02, f'bw={kde_bandwidth_pmotion:.4f}',
                ha='left', va='bottom', transform=axis_2.transAxes,
                fontsize='small',
                bbox=dict(boxstyle='round', facecolor='w', alpha=0.8))

    # Return stuff. Because I'm nice like that =)
    return [axis_1, axis_2]


def color_magnitude_diagram(fig,
                            axis,
                            data_gaia: pd.DataFrame,
                            cluster_labels: np.ndarray = None,
                            cluster_indices: Optional[Union[list, np.ndarray]] = None,
                            cluster_shading: Optional[np.ndarray] = None,
                            plot_std_limit: float = 3.0,
                            x_limits: Optional[Union[list, np.ndarray]] = None,
                            y_limits: Optional[Union[list, np.ndarray]] = None,
                            cluster_marker_radius: float = 1.0,
                            cmd_plot_color_key: bool = None):
    """Makes a colour magnitude diagram plot of a given star field, and may
    also overplot star clusters within that field.

    Notes:
        You can specify as few as one cluster_indices if cluster overplotting
        is too crowded.

    Args:
        fig (matplotlib figure): the figure element. Required to calculate alpha values precisely.
        axis (matplotlib axis): the colour magnitude diagram axis.
        data_gaia (pandas.DataFrame): the Gaia data read in to a DataFrame.
            Keys should be unchanged from default Gaia source table names.
        cluster_labels (np.ndarray): the cluster membership labels for
            clustered stars. This should be the default sklearn.cluster output.
        cluster_indices (list-like, optional): the clusters to plot in
            cluster_labels. For instance, you may not want to plot '-1'
            clusters (which are noise) produced by algorithm like DBSCAN.
        cluster_shading (np.ndarray, optional): an array of floats in the range 0, 1 for all clusters to use to shade
            their colour (aka alpha value). Useful for displaying e.g. HDBSCAN soft clustering.
            Default: None
        plot_std_limit (float): standard deviation of parameters to use
            to find plotting limits if none are explicitly specified.
            Default: 1.5
        x_limits (list-like, optional): the minimum and maximum blue minus red
            plotting limits.
        y_limits (list-like, optional): the minimum and maximum absolute
            magnitude plotting limits.
        cmd_plot_color_key (string, optional): optional other color in data_gaia to use instead of calculating G - RP.
        cluster_marker_radius (float): radius of the cluster marker. Useful to increase when clusters are hard to see
            against background points.
            Default: 1.0

    Returns:
        axis (matplotlib axis): the modified colour magnitude diagram axis.

    """

    # Calculate the stuff we need to plot
    # M = m - 5 * log10(d [pc]) + 5
    apparent_magnitude = data_gaia["phot_g_mean_mag"]

    if cmd_plot_color_key is None:
        color = (data_gaia["phot_bp_mean_mag"]
                 - data_gaia["phot_rp_mean_mag"])
    else:
        color = data_gaia[cmd_plot_color_key]

    # Calculate our own plotting limits if none have been specified
    if x_limits is None:
        x_limits = (np.mean(color)
                    + np.std(color)
                    * np.array([-plot_std_limit, plot_std_limit]))

    if y_limits is None:
        y_limits = (np.mean(apparent_magnitude)
                    + np.std(apparent_magnitude)
                    * np.array([-plot_std_limit, plot_std_limit]))

    if cluster_shading is None:
        cluster_shading = np.ones(data_gaia.shape[0])

    # Work out how many points will fall in the pm range
    good_x = np.logical_and(color > x_limits[0], color < x_limits[1])
    good_y = np.logical_and(apparent_magnitude > y_limits[0], apparent_magnitude < y_limits[1])
    n_points = np.count_nonzero(np.logical_and(good_x, good_y))

    # Calculate a rough guess at a good alpha value
    alpha_estimate = utilities.calculate_alpha(fig, axis, n_points, 1)

    # CMD plot
    axis.plot(color, apparent_magnitude,
              '.', ms=1, alpha=alpha_estimate, c='k')

    axis.set_xlabel(r'm_{bp} - m_{gp}')
    axis.set_ylabel(r'm_{G}')
    axis.set_title('colour magnitude diagram (Gaia photometry)')
    axis.set_xlim(x_limits)
    axis.set_ylim(y_limits)
    axis.invert_yaxis()

    # Plot clusters if labels have been specified
    if cluster_labels is not None:
        cmap = plt.get_cmap("tab10")  # Get the default matplotlib colourmap

        # We cycle over the clusters, grabbing their indices and plotting
        # them in a new colour erri time =)
        for i_color, a_cluster in enumerate(cluster_indices):
            stars_in_this_cluster = cluster_labels == a_cluster

            # Make colors and give them alpha values
            colors = np.tile(cmap(i_color % 10), (np.count_nonzero(stars_in_this_cluster), 1))
            colors[:, 3] = cluster_shading[stars_in_this_cluster]

            axis.scatter(color[stars_in_this_cluster],
                         apparent_magnitude[stars_in_this_cluster],
                         marker='.',
                         s=cluster_marker_radius**2,
                         c=colors,
                         zorder=100,
                         label=a_cluster)

    return axis


def ra_versus_parallax(fig,
                       axis,
                       data_gaia: pd.DataFrame,
                       cluster_labels: np.ndarray = None,
                       cluster_indices: Optional[Union[list, np.ndarray]] = None,
                       cluster_shading: Optional[np.ndarray] = None,
                       plot_std_limit: float = 3.0,
                       x_limits: Optional[Union[list, np.ndarray]] = None,
                       y_limits: Optional[Union[list, np.ndarray]] = None,
                       cluster_marker_radius: float = 1.0):
    """Makes a plot of right ascension vs parallax on an axis.

    Notes:
        You can specify as few as one cluster_indices if cluster overplotting
        is too crowded.

    Args:
        fig (matplotlib figure): the figure element. Required to calculate alpha values precisely.
        axis (matplotlib axis): the colour magnitude diagram axis.
        data_gaia (pandas.DataFrame): the Gaia data read in to a DataFrame.
            Keys should be unchanged from default Gaia source table names.
        cluster_labels (np.ndarray): the cluster membership labels for
            clustered stars. This should be the default sklearn.cluster output.
        cluster_indices (list-like, optional): the clusters to plot in
            cluster_labels. For instance, you may not want to plot '-1'
            clusters (which are noise) produced by algorithm like DBSCAN.
        cluster_shading (np.ndarray, optional): an array of floats in the range 0, 1 for all clusters to use to shade
            their colour (aka alpha value). Useful for displaying e.g. HDBSCAN soft clustering.
            Default: None
        plot_std_limit (float): standard deviation of parameters to use
            to find plotting limits if none are explicitly specified.
            Default: 1.5
        x_limits (list-like, optional): the minimum and maximum ra limits.
            Default: None
        y_limits (list-like, optional): the minimum and maximum parallax limits.
            Default: None
        cluster_marker_radius (float): radius of the cluster marker. Useful to increase when clusters are hard to see
            against background points.
            Default: 1.0

    Returns:
        axis (matplotlib axis): the modified colour magnitude diagram axis.

    """

    # Calculate the stuff we need to plot
    # M = m - 5 * log10(d [pc]) + 5
    ra = data_gaia['ra']
    parallax = data_gaia["parallax"]

    # Calculate our own plotting limits if none have been specified
    if x_limits is None:
        x_limits = np.asarray([ra.min(), ra.max()])

    if y_limits is None:
        y_limits = (np.mean(parallax)
                    + np.std(parallax)
                    * np.array([-plot_std_limit, plot_std_limit]))

    if cluster_shading is None:
        cluster_shading = np.ones(data_gaia.shape[0])

    # Work out how many points will fall in the pm range
    good_x = np.logical_and(ra > x_limits[0], ra < x_limits[1])
    good_y = np.logical_and(parallax > y_limits[0], parallax < y_limits[1])
    n_points = np.count_nonzero(np.logical_and(good_x, good_y))

    # Calculate a rough guess at a good alpha value
    alpha_estimate = utilities.calculate_alpha(fig, axis, n_points, 1)

    # CMD plot
    axis.plot(ra, parallax,
              '.', ms=1, alpha=alpha_estimate, c='k')

    axis.set_xlabel(r'ra')
    axis.set_ylabel(r'parallax')
    axis.set_title('parallax vs. ra')
    axis.set_xlim(x_limits)
    axis.set_ylim(y_limits)

    # Plot clusters if labels have been specified
    if cluster_labels is not None:
        cmap = plt.get_cmap("tab10")  # Get the default matplotlib colourmap

        # We cycle over the clusters, grabbing their indices and plotting
        # them in a new colour erri time =)
        for i_color, a_cluster in enumerate(cluster_indices):
            stars_in_this_cluster = cluster_labels == a_cluster

            # Make colors and give them alpha values
            colors = np.tile(cmap(i_color % 10), (np.count_nonzero(stars_in_this_cluster), 1))
            colors[:, 3] = cluster_shading[stars_in_this_cluster]

            axis.scatter(ra[stars_in_this_cluster],
                         parallax[stars_in_this_cluster],
                         marker='.',
                         s=cluster_marker_radius**2,
                         c=colors,
                         zorder=100,
                         label=a_cluster)

    return axis



# Todo: refactor to new structure (e.g. class-based, mirroring scipy.stats maybe?)

from typing import Union, Optional

import numpy as np


def king_surface_density(
    r_values: Union[float, np.ndarray],
    r_core: float,
    r_tidal: float,
    normalise: bool = False,
):
    """Computes the King surface density (King 1962, equation 14) given the three parameters. Can take vectorised input.
    Will return the surface density per square unit expected at a distance r_values from the core of the cluster.

    Valid only for:
        0 <= r < r_tidal (0 elsewhere - this is checked internally)
        0 < r_core < r_tidal (raises an error if not the case)

    Args:
        r_values (float or np.ndarray): r values to compute the surface density at.
        r_core (float): the core radius of the cluster.
        r_tidal (float): the tidal radius of the cluster.
        normalise (bool): whether or not to return the normalised King surface density profile, calculated
            numerically.
            Default: False

    Returns:
        a float or array of floats of the surface density for the cluster.

    """
    _check_core_and_tidal_radii(r_core, r_tidal)
    r_values = _check_r_values(r_values)

    # Constants
    rt_rc = r_tidal / r_core
    a = (1 + rt_rc**2) ** -0.5

    # Compute normalisation constant if desired
    if normalise:
        term_1 = r_core * np.arctan(rt_rc)
        term_2 = -2 * a * r_core * np.log(rt_rc + 1 / a)
        term_3 = r_tidal * a**2
        normalisation_constant = 1 / (term_1 + term_2 + term_3)
    else:
        normalisation_constant = 1.0

    # Work out where  0 <= r < r_tidal
    r_valid = np.logical_and(r_values >= 0, r_values < r_tidal)
    r_invalid = np.invert(r_valid)

    # Compute result
    reduced_r_values = (1 + (r_values[r_valid] / r_core) ** 2) ** (-0.5)
    result = np.empty(r_values.shape)
    result[r_valid] = normalisation_constant * (reduced_r_values - a) ** 2
    result[r_invalid] = 0

    return result


def _check_r_values(r_values):
    # Convert to a np array that's at least 1d, and check that it isn't bad
    r_values = np.atleast_1d(r_values)
    if not np.isfinite(r_values).all():
        raise ValueError("invalid r_values (e.g. nan or inf) are not allowed!")
    if np.any(r_values < 0):
        raise ValueError("all r_values must be positive or zero.")
    return r_values


def _check_core_and_tidal_radii(r_core, r_tidal):
    if not np.isfinite(r_core).all() or not np.isfinite(r_tidal).all():
        raise ValueError("input parameters must be finite!")
    if r_tidal < r_core or r_core < 0 or r_tidal < 0:
        raise ValueError("parameters must satisfy 0 < r_core < r_tidal")


def king_surface_density_fast(r_values: np.ndarray, r_core: float, r_tidal: float):
    """Computes the King surface density (King 1962, equation 14) given the three parameters. Can take vectorised input.
    Fast version intended for use with random sampling - it has no checks and cannot be normalised! Be careful!

    Valid only for:
        0 <= r < r_tidal (NOT CHECKED in this function!)
        0 < r_core < r_tidal (NOT CHECKED in this function!)

    Args:
        r_values (np.ndarray): r values to compute the surface density at. Must be a numpy array!
        r_core (float): the core radius of the cluster.
        r_tidal (float): the tidal radius of the cluster.

    Returns:
        a float or array of floats of the surface density for the cluster.

    """
    # Constants
    rt_rc = r_tidal / r_core
    a = (1 + rt_rc**2) ** -0.5

    # Compute result
    return ((1 + (r_values / r_core) ** 2) ** (-0.5) - a) ** 2


def sample_2d_king_profile(
    r_core: float,
    r_tidal: float,
    n_samples: int,
    seed=None,
    oversampling_factor: float = 10,
    return_generator: bool = False,
):
    """Samples a 2D King profile to return n_samples sample radii.

    Valid only for:
        0 < r_core < r_tidal

    Args:
        r_core (float): the core radius of the cluster.
        r_tidal (float): the tidal radius of the cluster.
        n_samples (int): the number of samples to generate.
        seed (int, optional): the seed of the random number generator. Default: None.
        oversampling_factor (float): how many times n_samples to generate each step, which helps to make sure that
            enough samples are quickly generated in just one or two loops. Default: 10.
        return_generator (bool): whether or not to return the numpy random number generator created. Default: False

    Returns:
        an array of sample radii of size n_samples, plus the random generator if return_generator==True.
    """
    _check_core_and_tidal_radii(r_core, r_tidal)

    r_samples = np.empty(n_samples, dtype=float)
    generator = np.random.default_rng(seed=seed)
    completed_samples = 0

    max_value = king_surface_density_fast(np.zeros(1), r_core, r_tidal)[0]

    while completed_samples < n_samples:
        remaining_samples = n_samples - completed_samples
        samples_to_generate = int(
            np.clip(remaining_samples * oversampling_factor, 10, np.inf)
        )

        # Generate some initial radius samples
        test_r_values = generator.uniform(high=r_tidal, size=samples_to_generate)
        test_king_values = king_surface_density_fast(test_r_values, r_core, r_tidal)
        test_mcmc_values = generator.uniform(size=samples_to_generate, high=max_value)

        # See which & how many are valid and save them!
        valid_test_samples = test_king_values < test_mcmc_values
        n_valid_test_samples = np.count_nonzero(valid_test_samples)

        if n_valid_test_samples > remaining_samples:
            r_samples[completed_samples:] = test_r_values[valid_test_samples][
                :remaining_samples
            ]
            break

        r_samples[
            completed_samples : completed_samples + n_valid_test_samples
        ] = test_r_values[valid_test_samples]
        completed_samples += n_valid_test_samples

    if return_generator:
        return r_samples, generator
    else:
        return r_samples


def sample_1d_king_profile(
    r_core: float,
    r_tidal: float,
    n_samples: int,
    seed: Optional[int] = None,
    oversampling_factor: float = 10,
):
    """Samples a 1D King profile (e.g. useful to get line of sight distances from the center of a cluster.) Uses a
    little trick - assumes that we're looking at the cluster side-on and removes the not-line-of-sight axis as if we
    were looking at it from the front. (This is because I cba to work out a 1D profile and more to the point, I couldn't
    fucking find one)

    Todo: change this to using the strip density g(x), which I *think* could do this better...

    Valid only for:
        0 < r_core < r_tidal

    Args:
        r_core (float): the core radius of the cluster.
        r_tidal (float): the tidal radius of the cluster.
        n_samples (int): the number of samples to generate.
        seed (int, optional): the seed of the random number generator. Default: None.
        oversampling_factor (float): how many times n_samples to generate each step, which helps to make sure that
            enough samples are quickly generated in just one or two loops. Default: 10.

    Returns:
        an array of sample radii of size n_samples
    """
    r_samples, generator = sample_2d_king_profile(
        r_core,
        r_tidal,
        n_samples,
        seed=seed,
        oversampling_factor=oversampling_factor,
        return_generator=True,
    )

    # Now, deproject this to 1D by giving every value an angle and then finding the 1D radius with the cosine
    random_angles = generator.uniform(high=2 * np.pi, size=n_samples)

    return r_samples * np.cos(random_angles)


def king_number_density(r, r_core, r_tidal, k=1):
    """Calculates the King1962 number density (eqn 18 in the paper.)
    
    Unnormalised by default (i.e. k=1.)
    """
    x = (r / r_core)**2
    x_t = (r_tidal / r_core)**2

    term_1 = np.log(1 + x)
    term_2 = - 4 * ((1 + x)**(0.5) - 1) / (1 + x_t)**(0.5)
    term_3 = x / (1 + x_t)
    
    return np.pi * r_core**2 * k * (term_1 + term_2 + term_3)




from . import epsilon
from .nearest_neighbor import precalculate_nn_distances
from .preprocess import cut_dataset, rescale_dataset, recenter_dataset
from .resample import generate_gaia_covariance_matrix, resample_gaia_astrometry
import warnings

warnings.warn(
    "ocelot.cluster API will change soon. Most objects will move, change, or be "
    "deprecated.",
    DeprecationWarning,
)



"""A set of functions for calculating optimum DBSCAN/OPTICS epsilon parameters of a field."""

import gc
from typing import Union, List, Tuple

import numpy as np
import pandas as pd

from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from .nearest_neighbor import precalculate_nn_distances


def castro_ginard(
    data_clustering: np.ndarray,
    nn_distances: np.ndarray,
    n_repeats: Union[int, List[int], Tuple[int]] = 10,
    min_samples: Union[str, int] = 10,
    return_std_deviation: bool = False,
    return_last_random_distance: bool = False,
):
    """A method for calculating an optimal epsilon value as in Alfred Castro-Ginard's 2018 paper (hence the acronym
    acg18.)

    Args:
        data_clustering (np.ndarray): clustering data for the field in shape (n_samples, n_features).
        nn_distances (np.ndarray): nearest neighbor distances for the field, in shape
            (n_samples, min_samples + 1).
        n_repeats (int, list of ints, tuple of ints): number of random repeats to perform. May be a list or tuple of
            integers.
            Default: 10
        min_samples (int, str): number of minimum samples to find the acg18 epsilon for (aka the kth nearest neighbor).
            May be an integer or 'all'.
            In this implementation, k = min_samples - 1 (k is the kth nearest neighbor to find.)
            Default: 10
        return_std_deviation (bool): whether or not to also return the standard deviation of epsilon estimates in the
            acg_epsilon dataframe.
            Default: False
        return_last_random_distance (bool): whether or not to return the final random distance set made by the function.
            Useful for making plots & understanding what's going on internally.
            Default: False

    Returns:
        - single float or np.ndarray of optimal epsilon values
        - if return_last_random_distance: an array of shape nn_distances.shape of the distances between stars in the
            last random distances simulation.

    """
    n_repeats = np.atleast_1d(n_repeats)
    max_n_repeats = np.max(n_repeats)

    if max_n_repeats < 1:
        raise ValueError("A positive number of repeats must be specified.")

    # Infer how many neighbors we need to calculate
    max_neighbors_to_calculate = nn_distances.shape[1]

    # Grab the minimum epsilon in the unperturbed field & do some error checking of min_samples in the process.
    if min_samples == "all":
        epsilon_minimum = np.min(nn_distances, axis=0)

    elif isinstance(min_samples, int):
        raise ValueError(
            "Incompatible number or string of min_samples specified.\n"
            "Allowed values:\n"
            "- integer less than / equal to max_neighbors_to_calculate + 1 and greater than zero\n"
            "- 'all', which calculates all values upto max_neighbors_to_calculate\n"
        )

    elif min_samples > max_neighbors_to_calculate + 1 or min_samples < 2:
        raise ValueError(
            "min_samples may not be larger than max_neighbors_to_calculate (aka nn_distances.shape[1]) + 1"
            " and must be a positive integer > 2."
        )

    else:
        kth_neighbor = min_samples - 1
        epsilon_minimum = np.min(
            nn_distances[:, kth_neighbor - 1]
        )  # -1 here because arrays are 0 indexed

    # Semi-paranoid memory management (lol)
    del nn_distances
    gc.collect()

    # Cycle over the required number of random repeats
    random_epsilons = np.zeros((max_n_repeats, max_neighbors_to_calculate))
    random_nn_distances = None  # Done solely to shut up my fucking linter
    i = 0
    while i < max_n_repeats:
        # Shuffle all values in the dataset column-wise, which is annoying to do but I guess I'll manage :L
        current_axis = 0
        while current_axis < data_clustering.shape[1]:
            data_clustering[:, current_axis] = np.random.permutation(
                data_clustering[:, current_axis]
            )
            current_axis += 1

        # Get some nn distances
        random_nn_distances = precalculate_nn_distances(
            data_clustering,
            n_neighbors=max_neighbors_to_calculate,
            return_sparse_matrix=False,
            return_knn_distance_array=True,
        )

        random_epsilons[i, :] = np.min(random_nn_distances, axis=0)

        i += 1

        # Semi-paranoid memory management (lol)
        if return_last_random_distance is False or i != max_n_repeats:
            del random_nn_distances
            gc.collect()

    # Ignore random epsilons we don't need if the user doesn't want them all
    if min_samples != "all":
        random_epsilons = random_epsilons[:, kth_neighbor - 1].reshape(-1, 1)
        acg_epsilon = {"min_samples": np.atleast_1d(min_samples)}
    else:
        acg_epsilon = {"min_samples": np.arange(max_neighbors_to_calculate) + 2}

    # Cycle over all the numbers of repeats the user requested, finding the requested acg epsilon values
    for a_n_repeats in n_repeats:
        mean_random_epsilons = np.mean(random_epsilons[:a_n_repeats, :], axis=0)
        acg_epsilon["acg_" + str(a_n_repeats)] = (
            mean_random_epsilons + epsilon_minimum
        ) / 2

        # Also add the standard deviation, if desired
        if return_std_deviation:
            acg_epsilon["acg_" + str(a_n_repeats) + "_std"] = np.std(
                random_epsilons[:a_n_repeats, :], axis=0
            )

    # Finally, turn this into a DataFrame!
    acg_epsilon_dataframe = pd.DataFrame(
        acg_epsilon, index=np.arange(len(acg_epsilon["min_samples"]))
    )

    if return_last_random_distance:
        return acg_epsilon_dataframe, random_nn_distances
    else:
        return acg_epsilon_dataframe


def kth_nn_distribution(r_range, a, dimension, k):
    """Returns the kth nearest neighbor distribution for a multi-dimensional ideal gas. Not normalised!

    f = r_range^(dimension + k - 1) / a^dimension * exp(-(r_range/a)^dimension)

    Args:
        r_range: radius values away from the center to evaluate at.
        a: the fitting constant
        k: the kth nearest neighbor moment of the distribution
        dimension: the assumed dimensionality of the distribution

    Returns:
        np.ndarray of the distribution evaluated at r_range

    """
    return (
        r_range ** (dimension + k - 1)
        / a**dimension
        * np.exp(-((r_range / a) ** dimension))
    )


def _summed_kth_nn_distribution_one_cluster(
    parameters: np.ndarray,
    k: int,
    r_range: np.ndarray,
    y_range: np.ndarray = None,
    minimisation_mode: bool = False,
):
    """Returns the summed kth nearest neighbor distribution, assuming the field contains at most one cluster.

    Todo: remove minimisation input_mode

    Args:
        parameters (np.ndarray): parameters of the model of length 5, in the form:
            0: field_constant (also known as a)
            1: field_dimension
            2: cluster_constant (also known as a)
            3: cluster_dimension
            4: cluster_fraction
        r_range (np.ndarray): radius values away from the center to evaluate at.
        k (int): the kth nearest neighbor moment of the distribution
        y_range (np.ndarray): unsupported previous functionality. Raises error if not None.
        minimisation_mode (bool): unsupported previous functionality. Raises error if not False.

    Returns:
        an array of y_field, y_cluster and y_total.

    """
    if minimisation_mode or y_range is not None:
        raise NotImplementedError(
            "This function no longer supports use with previous minimisation versions of the "
            "field model for epsilon determination."
        )

    # Calculate cumulatively summed (and normalised) distributions for both the field and the cluster
    y_field = np.cumsum(kth_nn_distribution(r_range, parameters[0], parameters[1], k))
    normalisation_field = np.trapz(y_field, x=r_range) / (1 - parameters[4])

    y_cluster = np.cumsum(kth_nn_distribution(r_range, parameters[2], parameters[3], k))
    normalisation_cluster = np.trapz(y_cluster, x=r_range) / parameters[4]

    y_field /= normalisation_field
    y_cluster /= normalisation_cluster
    y_total = y_field + y_cluster

    # Return raw values to be used by a plotter (slow due to the initialisation process, amongst other things)
    # Make a big array to work on
    log_array = np.vstack([y_field, y_cluster, y_total])
    good_values = log_array > 0

    # Take logs only where log() is defined, otherwise replace with -np.inf
    log_array = np.where(good_values, np.log10(log_array, where=good_values), -np.inf)

    return log_array


def constrained_a(d, k, epsilon_max):
    # Todo I think this can just be a part of the class _SummedKNNOneClusterCurveFit
    return epsilon_max / (((k - 1) / d + 1) ** (1 / d))


class _SummedKNNOneClusterCurveFit:
    def __init__(self, k: int, epsilon_max: float):
        """Class to handle kth nearest neighbor curve fits. Unchanging parameters are kept stored by this class, and
        the __call__ method allows for access to the kth nn curve fitting.

               epsilon max
                   |
        density    v
        ^          _
        |         / \
        |        /   `
        |   _   /       `  _
        |  / |_/             `
        |____________________>
        kth nearest neighbour distance

        Args:
            k (int): the order of the knn plot. E.g. k=4 corresponds to a 4th nearest neighbor fit.
            epsilon_max (float): the maximum value of an epsilon density histogram for the kth nearest neighbor.

        """
        self.k = k
        self.epsilon_max = epsilon_max

    def __call__(
        self,
        r_range: np.ndarray,
        *parameters,
    ) -> np.ndarray:
        """Accesses the kth nearest neighbor distribution curve fitting routine in a fast way.

        Args:
            r_range (np.ndarray): x points/epsilon points to evaluate at, with shape (n_samples,).
            *parameters: parameters of the distribution. They should be in the order:
                0: field_dimension
                1: cluster_maximum
                2: cluster_dimension
                3: cluster_fraction

        Returns:
            the kth nearest neighbor distribution y values (aka log point number values) in an array of shape
            (n_samples,).

        """
        # Calculate the as
        a_field = constrained_a(parameters[0], self.k, self.epsilon_max)
        a_cluster = constrained_a(parameters[2], self.k, parameters[1])

        # Calculate cumulatively summed (and normalised) distributions for both the field and the cluster
        y_field = np.cumsum(
            kth_nn_distribution(r_range, a_field, parameters[0], self.k)
        )
        normalisation_field = np.trapz(y_field, x=r_range) / (1 - parameters[3])

        y_cluster = np.cumsum(
            kth_nn_distribution(r_range, a_cluster, parameters[2], self.k)
        )
        normalisation_cluster = np.trapz(y_cluster, x=r_range) / parameters[3]

        y_field /= normalisation_field
        y_cluster /= normalisation_cluster
        y_total = y_field + y_cluster

        return np.log10(y_total)


def _get_epsilon_plotting_styles(epsilon_values):
    """Quick function to neaten my code and get default plotting styles for epsilon values for a diagnostic plot.

    Args:
        epsilon_values (list-like): epsilon values to plot, in the usual order (0, 1, 2)

    Returns:
        a new functions_to_overplot list

    """
    return [
        {
            "label": f"eps_c: {epsilon_values[0]:.4f}",
            "style": "r:",
            "x": [epsilon_values[0]] * 2,
            "y": [1e-300, 1e300],
            "differentiate": False,
        },
        {
            "label": f"eps_n1: {epsilon_values[1]:.4f}",
            "style": "k:",
            "x": [epsilon_values[1]] * 2,
            "y": [1e-300, 1e300],
            "differentiate": False,
        },
        {
            "label": f"eps_n2: {epsilon_values[2]:.4f}",
            "style": "k-.",
            "x": [epsilon_values[2]] * 2,
            "y": [1e-300, 1e300],
            "differentiate": False,
        },
        {
            "label": f"eps_n3: {epsilon_values[3]:.4f}",
            "style": "k:",
            "x": [epsilon_values[3]] * 2,
            "y": [1e-300, 1e300],
            "differentiate": False,
        },
        {
            "label": f"eps_f: {epsilon_values[4]:.4f}",
            "style": "r:",
            "x": [epsilon_values[4]] * 2,
            "y": [1e-300, 1e300],
            "differentiate": False,
        },
    ]


def _get_model_plotting_styles(x_range, y_field, y_cluster, y_total):
    """Quick function to neaten my code and get default plotting styles for epsilon values for a diagnostic plot.

    Args:
        x_range, y_field, y_cluster, y_total: self-explanatory tbh. Everyone shares an x_range!

    Returns:
        a new functions_to_overplot list

    """
    return [
        {
            "label": "field model",
            "style": "m--",
            "x": x_range,
            "y": 10**y_field,
            "differentiate": True,
        },
        {
            "label": "cluster model",
            "style": "c--",
            "x": x_range,
            "y": 10**y_cluster,
            "differentiate": True,
        },
        {
            "label": "total model",
            "style": "r-",
            "x": x_range,
            "y": 10**y_total,
            "differentiate": True,
        },
    ]


def _find_curve_absolute_maximum_epsilons(
    x_range: np.ndarray,
    y_range: np.ndarray,
):
    """Returns the epsilon value corresponding to the absolute maximum value of the second derivative of a curve.
    Useful for finding the characteristic epsilon value the cluster, e_c.

    Args:
        x_range (np.ndarray): x values of the calculated fit.
        y_range (np.ndarray): y values of the calculated fit.

    Returns:
        the x/epsilon value corresponding to the maximum of the second y derivative, ignoring any points at which the
            second derivative was infinite/NaN.
    """
    # Take the second derivative, albeit only on good stars
    good_stars = np.logical_and(
        np.logical_and(np.isfinite(x_range), x_range > 0), np.isfinite(y_range)
    )

    x_range = np.log10(x_range[good_stars])
    y_range = y_range[good_stars]

    d2_y_range = np.gradient(np.gradient(y_range, x_range), x_range)

    # Take absolute values, but only when the values are finite just to be super safe with the input. When not finite,
    # we set it to -1 (reflecting that it's definitely not the maximum value we want.)
    good_stars = np.isfinite(d2_y_range)
    bad_stars = np.invert(good_stars)

    d2_y_range_absolute = np.zeros(d2_y_range.shape)
    d2_y_range_absolute[good_stars] = np.abs(d2_y_range[good_stars])
    d2_y_range_absolute[bad_stars] = -1.0

    # Find the maximum and return!
    d2_y_range_max_id = np.argmax(d2_y_range_absolute)
    return 10 ** x_range[d2_y_range_max_id]


def _find_sign_change_epsilons(
    x_range: np.ndarray, y_range: np.ndarray, return_all_sign_changes: bool = False
):
    """Takes derivatives of a numerical cluster model to find the beginning, middle and end of the area with the
    steepest change in gradient, which typically corresponds to the point at which field stars begin to dominate.

    log point number
        ^            ________
        |           /
        |    ______/
        |   /     ^^^
        |  / eps: 012
        |____________________>
    kth nearest neighbour distance

    Notes:
        - DO NOT CALL THIS FUNCTION if cluster_fraction is extremely low: it may or may not find an epsilon value (from
            floating point errors) and that value may or may not be completely stupid. Be careful!
        - In initial development, I raised x_range to the power of ten to weight peaks further from epsilon=0 more
            highly. Once I switched to using models, this wasn't necessary, as the derivative was comparatively much
            more stable. However, that might just be a function of later testing. Worth bearing in mind if it suddenly
            doesn't work well. I think this was especially an issue if the first few points were noisy (and hence had an
            anomalously high area under their 2nd derivative.)
        - 20/01/28 follow-up to the above: I'm now actually logging x_range. There was a small numerical difference
            between what was visible on plots in log space and what this function returned. Pet theory: log space
            makes the derivative much less abrupt and smoother, making this function's job easier. The changes in
            the function show up more clearly & smoothly in log space.

    Args:
        x_range (np.ndarray): x values of the calculated fit.
        y_range (np.ndarray): y values of the calculated fit.
        return_all_sign_changes (bool): whether or not to also return a list of all sign changes found. Useful for
            debugging.
            Default: False

    Returns:
        - a list containing epsilon 0, 1 and 2
        - return_all_sign_changes=True: also returns an array of all sign changes found

    """
    # Take the second derivative, albeit only on good stars
    good_stars = np.logical_and(
        np.logical_and(np.isfinite(x_range), x_range > 0), np.isfinite(y_range)
    )

    x_range = np.log10(x_range[good_stars])
    y_range = y_range[good_stars]

    d2_y_range = np.gradient(np.gradient(y_range, x_range), x_range)

    # Find & output some cool values
    # Firstly, find all candidate 2nd derivative = 0 points (epsilon 0 or 2) by looking for sign changes, based on
    # https://stackoverflow.com/questions/2652368/how-to-detect-a-sign-change-for-elements-in-a-numpy-array
    sign_change_ids = (
        np.asarray(np.sign(d2_y_range[:-1]) != np.sign(d2_y_range[1:])).nonzero()[0] + 1
    )
    sign_change_x_values = x_range[sign_change_ids]

    # We make a clipped d2_y_range so that anything under the curve is ignored
    clipped_d2_y_range = np.clip(d2_y_range, 0, np.inf)

    # Cycle over pairs and record which one has the highest area below it
    i = 0
    epsilon_0_and_2_ids = np.zeros(2, dtype=int)
    epsilon_0_and_2_x_values = np.zeros(2)
    max_area = -1.0

    while i < sign_change_x_values.shape[0] - 1:
        id_0 = sign_change_ids[i]
        id_1 = sign_change_ids[i + 1]

        area = np.trapz(clipped_d2_y_range[id_0 : id_1 + 1], x=x_range[id_0 : id_1 + 1])

        # Save this pair if it's the best
        if area > max_area:
            epsilon_0_and_2_ids = np.array([id_0, id_1])
            epsilon_0_and_2_x_values = sign_change_x_values[[i, i + 1]]
            max_area = area

        i += 1

    # Next, find the point of maximum curvature between eps0 and eps2 by finding the maximum value of the 2nd derivative
    # But we only do that if there were even any sign changes ( => i == 0 since no incrementation has happened)
    if i == 0:
        epsilon_1_x_value = 0.0
    else:
        epsilon_1_id = np.argmax(
            d2_y_range[epsilon_0_and_2_ids[0] : epsilon_0_and_2_ids[1] + 1]
        )
        epsilon_1_x_value = (
            x_range[epsilon_0_and_2_ids[0] : epsilon_0_and_2_ids[1] + 1]
        )[epsilon_1_id]

    # Unlog the answers
    epsilon_1_x_value = 10**epsilon_1_x_value
    epsilon_0_and_2_x_values = 10**epsilon_0_and_2_x_values
    sign_change_x_values = 10**sign_change_x_values

    # Happy return time =)
    if return_all_sign_changes:
        return [
            epsilon_0_and_2_x_values[0],
            epsilon_1_x_value,
            epsilon_0_and_2_x_values[1],
        ], sign_change_x_values
    else:
        return [
            epsilon_0_and_2_x_values[0],
            epsilon_1_x_value,
            epsilon_0_and_2_x_values[1],
        ]


def field_model(
    nn_distances: np.ndarray,
    min_samples: int = 10,
    min_cluster_size: int = 1,
    resolution: int = 500,
    point_fraction_to_keep: float = 0.95,
    model_minimum_drop: float = 1.0,
    optimiser: str = "trf",
    print_convergence_messages: bool = False,
    make_diagnostic_plot: bool = False,
    return_all_sign_changes: bool = False,
    **kwargs,
) -> list:
    """Attempts to find an optimum value for epsilon by modelling the field of the cluster and the cluster itself.
    Leverages scipy curve fitting to find optimum model values, and can even report on the approximate estimated size
    of a cluster in the given field.

    Differs from the normal field_model_cf in that the maximum value of the field model is constrained by the dataset. A
    histogram is made (and its max value found) of all the epsilons. The field's a parameter is then fixed to this,
    using the equation:

    a = epsilon_max / ( (k - 1) / d + 1 )^(1/d)

    Args:
        nn_distances (np.ndarray): nearest neighbor distances for the field, in shape
            (n_samples, max_neighbors_to_calculate).
        min_samples (int): number of minimum samples to find the epsilon for (aka the kth nearest neighbor).
            Default: 10
        min_cluster_size (int): minimum allowed size of a cluster, based on the value of cluster_fraction derived in
            the fitting procedure. Setting this larger can help to avoid high epsilons that return noise clusters.
            Default: 1 (virtually equivalent to setting this to off)
        resolution (int): resolution to re-sample the data to. Should be high enough that all detail is kept, but not
            so high as to drastically slow down the program.
            Default: 500
        point_fraction_to_keep (float): for efficiency reasons, points with a very high epsilon should be dropped. This
            makes re-sampling the data require far fewer points and ensures the minimiser will focus more on the cluster
            (at low epsilon.) The bottom point_fraction_to_keep fraction of points is kept.
            Default: 0.95 (i.e. 5% of points with the highest epsilon are removed, a good general value)
        model_minimum_drop (float): a number subtracted from the minimum input log normalised point number that defines
            how low the total model is sampled. You may want to use a number like 1.0 if turnoffs or details are only
            just being missed.
            Default: 0.0, i.e. we'll have the same minimum point value in the model as the input.
        optimiser (string): optimiser to be used by scipy.optimize.curve_fit. trf is great, and lets bounds be used.
            Default: 'trf'
        print_convergence_messages (bool): whether or not to ask scipy.optimize.curve_fit to print convergence messages.
            Default: False
        make_diagnostic_plot (bool): whether or not to make a diagnostic plot with
            ocelot.plot.nearest_neighbor_distances of the results we've got.
            Default: False
        return_all_sign_changes (bool): whether or not to also return a list of all sign changes found. Useful for
            debugging.
            Default: False
        **kwargs: keyword arguments to pass to ocelot.plot.nearest_neighbor_distances

    Returns:
        - bool for whether or not a cluster was found
        - a tuple containing epsilon 0, 1, 2, 3 and 4 estimates
        - a tuple of the fitting parameters found
        - the expected number of cluster members, n_cluster_members

    """
    # -- Pre-processing
    # Grab the correct neighbor distances, sort them and drop stuff we don't want
    distances = np.sort(nn_distances[:, min_samples - 1])
    distances = distances[: int(point_fraction_to_keep * distances.shape[0])]

    # Create a normalised log number of points array
    points = np.arange(1, distances.shape[0] + 1)
    points = points / np.trapz(points, x=distances)
    points = np.log10(points)

    # Interpolate it to ensure the points are linearly sampled and reduce noise
    interpolator = interp1d(distances, points, kind="linear")
    distances_interpolated = np.linspace(
        distances.min(), distances.max(), num=resolution
    )
    points_interpolated = interpolator(distances_interpolated)

    # -- Get maximum value of epsilon
    # Histogram all the distances
    bin_values, bin_edges = np.histogram(distances, bins="auto")

    # Grab the max bin
    max_bin = np.argmax(bin_values)
    modal_epsilon_value = (bin_edges[max_bin] + bin_edges[max_bin + 1]) / 2

    # -- Fitting
    # Define bounds       dim_f   max_cluster          dim_c   cluster_frac
    bounds = (
        np.asarray([2.0, distances.min(), 2.0, 0.0]),
        np.asarray([np.inf, modal_epsilon_value, np.inf, 1.0]),
    )

    # Grab an initial guess
    field_dimension = 5
    cluster_maximum = np.clip(
        modal_epsilon_value / 2, distances.min(), modal_epsilon_value
    )
    cluster_dimension = 3
    cluster_fraction = 0.01

    # Minimisation time! Parameters is the stuff to minimise, arguments is the stuff we pass to the function to use it
    parameters = np.asarray(
        [field_dimension, cluster_maximum, cluster_dimension, cluster_fraction]
    )

    curve_to_fit = _SummedKNNOneClusterCurveFit(min_samples, modal_epsilon_value)

    result_unprocessed, covariance = curve_fit(
        curve_to_fit,
        distances_interpolated,
        points_interpolated,
        p0=parameters,
        bounds=bounds,
        method=optimiser,
        verbose=print_convergence_messages,
    )

    # Process the result array back into the array of parameters we like to see around these parts
    # Since result unprocessed is [field_dimension, cluster_maximum , cluster_dimension, cluster_fraction]
    # but we want [field_constant, field_dimension, cluster_constant, cluster_dimension, cluster_fraction]
    field_constant = constrained_a(
        result_unprocessed[0], min_samples, modal_epsilon_value
    )
    cluster_constant = constrained_a(
        result_unprocessed[2], min_samples, result_unprocessed[1]
    )
    result = np.asarray(
        [
            field_constant,
            result_unprocessed[0],
            cluster_constant,
            result_unprocessed[2],
            result_unprocessed[3],
        ]
    )

    # -- Calculation and grabbing of epsilon values
    # We make sure that we oversample the function from near-zero, which helps to find all points where it crosses axis
    distances_sampled = np.linspace(1e-30, distances.max(), num=resolution)

    # We'll evaluate the function at the results that were grabbed, for plotting and epsilon purposes
    (
        points_field,
        points_cluster,
        points_total,
    ) = _summed_kth_nn_distribution_one_cluster(
        result, min_samples, distances_sampled, minimisation_mode=False
    )

    # And we'll only keep valid points for which the total model does not have a point number value less than that
    # of model_minimum_drop * min(points)
    good_values = points_total >= np.min(points) - model_minimum_drop

    points_total = points_total[good_values]
    points_cluster = points_cluster[good_values]
    points_field = points_field[good_values]
    distances_sampled = distances_sampled[good_values]

    # Calculate the estimated number of cluster members and calculate epsilon if it's equal or above the minimum size
    n_cluster_members = int(np.round(distances.shape[0] * result[-1]))
    if n_cluster_members >= min_cluster_size:
        epsilon_values = np.zeros(5)

        epsilon_values[0] = _find_curve_absolute_maximum_epsilons(
            distances_sampled, points_cluster
        )
        epsilon_values[1:4], sign_changes = _find_sign_change_epsilons(
            distances_sampled, points_total, return_all_sign_changes=True
        )
        epsilon_values[4] = modal_epsilon_value

        functions_to_overplot_epsilon_values = _get_epsilon_plotting_styles(
            epsilon_values
        )
    else:
        epsilon_values = sign_changes = None
        functions_to_overplot_epsilon_values = []

    # -- Diagnostic plotting
    # Diagnostic plot if we're asked nicely
    if make_diagnostic_plot:
        raise NotImplementedError(
            "Diagnostic plotting was removed in a previous version."
        )

    # Return nada if no cluster was found
    if n_cluster_members < min_cluster_size:
        to_return = [False, (0.0, 0.0, 0.0, 0.0, 0.0), result, n_cluster_members]
    else:
        to_return = [True, epsilon_values, result, n_cluster_members]

    if return_all_sign_changes:
        to_return += sign_changes

    return to_return



"""A set of tests for use with the pytest module, covering ocelot.isochrone"""

# FUCKING HATE PYTHON IMPORTS AAAA
# (the below fixes this though)
try:
    from .context import ocelot
except ModuleNotFoundError:
    print('Unable to find ocelot via .context! Trying to import from your python path instead...')
try:
    import ocelot
except ModuleNotFoundError:
    raise ModuleNotFoundError('Unable to find ocelot')

import pickle
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import matplotlib.pyplot as plt
from scipy.sparse.csr import csr_matrix

path_to_blanco_1 = Path('./test_data/blanco_1_gaia_dr2_gmag_18_cut.pickle')
path_to_healpix_pixel = Path('./test_data/healpix_12237.csv')
path_to_healpix_pixels = Path('./test_data/healpix_pixel/')

path_to_one_simulated_population = Path('./test_data/simulated_populations/small/1.dat')
path_to_all_simulated_populations = Path('./test_data/simulated_populations/small')
path_to_big_simulated_population = Path('./test_data/simulated_populations/large/2.dat')

path_to_simulated_population_test_clusters = Path('./test_data/simulated_populations/test_clusters')


def test_cut_dataset():
    """Tests the functionality of dataset cutting at ocelot.cluster.cut_dataset."""
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Define some chonky cuts to make
    cuts = {'phot_g_mean_mag': [6, 16],
            'r_est': [100, np.inf],
            'parallax_over_error': [10, np.inf]}

    data_cut = ocelot.cluster.cut_dataset(data_gaia, parameter_cuts=cuts)

    # Check that the correct number of cuts were performed
    assert data_cut.shape == (5039, 28)

    # Check that no values fall outside of the cut ranges for gmag (easy)
    assert np.invert(np.any(np.logical_or(data_cut['phot_g_mean_mag'] < cuts['phot_g_mean_mag'][0],
                                          data_cut['phot_g_mean_mag'] > cuts['phot_g_mean_mag'][1])))

    # Check that no values fall outside of the cut ranges for r_est (could indicate fucking up with np.inf)
    assert np.invert(np.any(np.logical_or(data_cut['r_est'] < cuts['r_est'][0],
                                          data_cut['r_est'] > cuts['r_est'][1])))

    return data_cut


def _plot_for_recenter_dataset(data_gaia: pd.DataFrame, super_title: str = ''):
    """Cheeky plotting function for manual testing of the dataset recentering functionality."""
    fig, ax = plt.subplots(nrows=2, ncols=2, figsize=(8, 8))

    marker_radius = 1
    alpha = 0.3

    ax[0, 0].scatter(data_gaia['ra'], data_gaia['dec'], s=marker_radius**2, alpha=alpha)
    ax[0, 1].scatter(data_gaia['pmra'], data_gaia['pmdec'], s=marker_radius**2, alpha=alpha)
    ax[1, 0].scatter(data_gaia['lon'], data_gaia['lat'], s=marker_radius**2, alpha=alpha)
    ax[1, 1].scatter(data_gaia['pmlon'], data_gaia['pmlat'], s=marker_radius**2, alpha=alpha)

    ax[0, 0].set_title('ra vs dec')
    ax[0, 1].set_title('pmra vs pmdec')
    ax[1, 0].set_title('lon vs lat')
    ax[1, 1].set_title('pmlon vs pmlat')

    fig.suptitle(super_title)

    fig.show()
    plt.close(fig)


def test_recenter_dataset(show_figure=False):
    """Tests that position recentering in ocelot.cluster.recenter_dataset works as intended."""
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    center = [data_gaia['ra'].median(), data_gaia['dec'].median()]

    data_gaia = ocelot.cluster.recenter_dataset(data_gaia, center=center)

    if show_figure:
        _plot_for_recenter_dataset(data_gaia, super_title="test_recenter_dataset")

    # A quick check that the median values are about 0, 0 (won't be exact due to distortions)
    assert np.allclose(0.0, data_gaia['lon'].median(), rtol=0.0, atol=0.05)
    assert np.allclose(0.0, data_gaia['lat'].median(), rtol=0.0, atol=0.05)

    return data_gaia


def test_recenter_dataset_healpix(show_figure=False):
    """Tests that position recentering in ocelot.cluster.recenter_dataset works as intended, but for when using a
    healpix pixel."""
    # Read in data for the pixel
    data_gaia = pd.read_csv(path_to_healpix_pixel)

    data_gaia = ocelot.cluster.recenter_dataset(data_gaia, pixel_id=12237, rotate_frame=True)

    if show_figure:
        _plot_for_recenter_dataset(data_gaia, super_title="test_recenter_dataset_healpix")

    # A quick check that the median values are about 0, 0 (won't be exact due to distortions)
    assert np.allclose(0.0, data_gaia['lon'].median(), rtol=0.0, atol=0.05)
    assert np.allclose(0.0, data_gaia['lat'].median(), rtol=0.0, atol=0.05)

    return data_gaia


def test_rescale_dataset():
    """Tests the functionality of the dataset re-scaling of ocelot.cluster.rescale_dataset."""
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Re-scale the data with standard scaling and check that it has zero mean, unit variance
    data_rescaled, data_rescaled_2 = ocelot.cluster.rescale_dataset(data_gaia, data_gaia, scaling_type='standard',
                                                                    concatenate=False)
    data_rescaled_3 = ocelot.cluster.rescale_dataset(data_gaia, scaling_type='standard')

    # Check for zero mean, unit standard deviation
    means = np.mean(data_rescaled, axis=0)
    assert np.allclose(means, np.zeros(5), rtol=0.0, atol=1e-8)

    std_deviation = np.std(data_rescaled, axis=0)
    assert np.allclose(std_deviation, np.ones(5), rtol=0.0, atol=1e-8)

    # Check that the additional data frame was correctly re-scaled too - they should be identical
    assert np.allclose(data_rescaled, data_rescaled_2, rtol=1e-6, atol=1e-8)

    # And check that doing it just once has the same result too
    assert np.allclose(data_rescaled, data_rescaled_3, rtol=1e-6, atol=1e-8)

    # Check that calling an invalid scaler type throws an error
    with pytest.raises(ValueError, match="Selected scaling_type not supported!"):
        ocelot.cluster.rescale_dataset(data_gaia, scaling_type='an unsupported type of scaling, I guess')

    # Check that a cheeky nan value raises an error
    data_gaia.loc[100, 'ra'] = np.nan
    with pytest.raises(ValueError, match="At least one value in data_gaia is not finite! Unable to rescale the data."):
        ocelot.cluster.rescale_dataset(data_gaia, scaling_type='robust')

    return data_gaia, data_rescaled


def test_precalculate_nn_distances():
    """Tests ocelot.cluster.precalculate_nn_distances()"""
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Re-scale the data with standard scaling and check that it has zero mean, unit variance
    data_rescaled = ocelot.cluster.rescale_dataset(data_gaia, scaling_type='robust')

    # Calculate some nearest neighbor distances baby!
    sparse_matrix, distance_matrix = ocelot.cluster.precalculate_nn_distances(
        data_rescaled, n_neighbors=10, return_sparse_matrix=True, return_knn_distance_array=True)

    # Check that specifying no return type throws an error
    with pytest.raises(ValueError, match="Nothing was specified for return. That's probably not intentional!"):
        ocelot.cluster.precalculate_nn_distances(data_rescaled, n_neighbors=10,
                                                 return_sparse_matrix=False, return_knn_distance_array=False)

    # Check that the sparse matrix is, in fact, sparse
    assert type(sparse_matrix) == csr_matrix

    # Check that the correct shapes are returned
    assert distance_matrix.shape == (14785, 10)
    assert sparse_matrix.shape == (14785, 14785)

    return sparse_matrix, distance_matrix


def test_find_largest_cluster():
    """Tests ocelot.cluster.partition.find_largest_cluster_in_pixel."""
    largest_cluster = ocelot.cluster.partition.find_largest_cluster_in_pixel([0], 5, 1)

    assert np.allclose(largest_cluster, 15.886934673445522, atol=1e-3, rtol=1e-3)

    return largest_cluster


def test_data_partition(show_figure=False):
    """Tests all elements of the data partition class, ocelot.cluster.DataPartition."""
    # Read in the data
    to_concatenate = []
    for a_file in path_to_healpix_pixels.glob('*.csv'):
        to_concatenate.append(pd.read_csv(a_file))

    data_gaia = pd.concat(to_concatenate, ignore_index=True)
    del to_concatenate
    
    # We want the data to have lon, lat values for our upcoming unit test use
    data_gaia = ocelot.cluster.recenter_dataset(data_gaia, pixel_id=12238)

    # Define what our partition will look like
    constraints = [
        [None, None, 0],
        [5, 7, 700],
        [6, 8, 3000],
    ]
    
    size_threshold = 4

    # Let's go!
    partitioner = ocelot.cluster.DataPartition(data_gaia,
                                               12238,
                                               constraints=constraints,
                                               final_distance=np.inf,
                                               parallax_sigma_threshold=2.,
                                               minimum_size=size_threshold,
                                               n_stars_per_component=[800, 500, 200],
                                               verbose=True)

    # Check the total number of partitions
    # assert partitioner.total_partitions == 11

    # Test that we can check whether or not something is internal to the partitions properly
    # First off, check a single star in the first partition (easy)
    partitioner.get_partition(0)
    assert partitioner.test_if_in_current_partition(0, 0, 20)

    # Check that getting of n_components works
    assert partitioner.get_n_components() == 11
    
    # Now, let's check some more values. The very last one is the only one with a bad parallax.
    lons = np.asarray([-0.608355,  0.991817,  1.991817, 0.581063, 2.3,  0.])
    lats = np.asarray([-2.737361, -0.958997, -0.958997, -2.1,     2.2, -2.])
    parallaxes = np.asarray([1.3, 1.2, 1.1, 1.0, 0.9, 1.7])

    partitioner.get_partition(8)
    target_array = np.asarray([True, True, False, True, False, False])
    result_array = partitioner.test_if_in_current_partition(lons, lats, parallaxes)

    assert np.all(target_array == result_array)

    # Run the plotting code (integration test alert!!!! Call the bad testing police on me why don't you)
    partitioner.plot_partition_bar_chart(figure_title='test of the partitioner', show_figure=show_figure,
                                         desired_size=size_threshold, base_n_stars_per_component=800)
    partitioner.plot_partitions(figure_title='testing of the partitioner', show_figure=show_figure,
                                cmd_plot_y_limits=[9, 16])

    return data_gaia, partitioner


def test_acg18():
    """Tests the OCELOT implementation of the Alfred Castro-Ginard+18 method of determining an optimum value for DBSCAN.
    """
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Define some chonky cuts to make to make this test faster
    cuts = {'phot_g_mean_mag': [-np.inf, 16],
            'r_est': [200, 300],
            'parallax_over_error': [5, np.inf]}
    cuts = {}

    data_cut = ocelot.cluster.cut_dataset(data_gaia, parameter_cuts=cuts)

    # Re-scale the data with standard scaling and check that it has zero mean, unit variance
    data_rescaled = ocelot.cluster.rescale_dataset(data_cut, scaling_type='robust')

    # Calculate some nearest neighbor distances baby!
    distance_matrix = ocelot.cluster.precalculate_nn_distances(
        data_rescaled, n_neighbors=10, return_sparse_matrix=False, return_knn_distance_array=True)

    # Epsilon time
    np.random.seed(42)
    epsilons, random_distance_matrix = ocelot.cluster.epsilon.acg18(
        data_rescaled, distance_matrix, n_repeats=[1, 2], min_samples='all', return_last_random_distance=True,
        return_std_deviation=True)

    # Check that the correct shapes are returned
    assert distance_matrix.shape == random_distance_matrix.shape
    assert epsilons.shape == (10, 5)
    assert epsilons.columns.to_list() == ['min_samples', 'acg_1', 'acg_1_std', 'acg_2', 'acg_2_std']

    # Check the values against a target set that were correct at first implementation
    target_epsilons = np.asarray([[0.01767441, 0.01594943],
                                  [0.06125835, 0.06773323],
                                  [0.09002014, 0.09449874],
                                  [0.10619648, 0.10952761],
                                  [0.12325615, 0.12000926],
                                  [0.12975783, 0.13023932],
                                  [0.14481370, 0.14131321],
                                  [0.14834746, 0.14809204],
                                  [0.15089530, 0.15114417],
                                  [0.15616676, 0.15699153]])
    assert np.allclose(epsilons[['acg_1', 'acg_2']].values, target_epsilons, rtol=0.0, atol=1e-8)
    
    # Also check the standard deviations
    target_epsilons = np.asarray([[0., 0.00344996],
                                  [0., 0.01294977],
                                  [0., 0.00895719],
                                  [0., 0.00666226],
                                  [0., 0.00649378],
                                  [0., 0.00096298],
                                  [0., 0.00700098],
                                  [0., 0.00051084],
                                  [0., 0.00049773],
                                  [0., 0.00164952]])
    assert np.allclose(epsilons[['acg_1_std', 'acg_2_std']].values, target_epsilons, rtol=0.0, atol=1e-8)
    
    # And check the range of min_samples. It should go from 2 to 11 when provided with values of k from 0 to 10.
    assert np.all(epsilons['min_samples'].values == np.arange(10) + 2)

    # Check that we just get back one (correct) value if that's all we ask for
    np.random.seed(42)
    single_epsilon = ocelot.cluster.epsilon.acg18(
        data_rescaled, distance_matrix, n_repeats=2, min_samples=10, return_last_random_distance=False)
    single_epsilon = single_epsilon['acg_2'].values[0]

    assert type(single_epsilon) == float or np.float
    assert np.allclose(single_epsilon, 0.15144678787864096, rtol=0.0, atol=1e-8)

    # Throw some errors by being a fuckwit with min_samples
    # Wrong string
    with pytest.raises(ValueError, match="Incompatible number or string of min_samples specified."):
        ocelot.cluster.epsilon.acg18(
            data_rescaled, distance_matrix, n_repeats=2, min_samples='not the droid you were looking for', 
            return_last_random_distance=False)

    # Min_samples greater than the number of neighbours
    with pytest.raises(ValueError,
                       match="min_samples may not be larger than max_neighbors_to_calculate"):
        ocelot.cluster.epsilon.acg18(
            data_rescaled, distance_matrix, n_repeats=2, min_samples=100, return_last_random_distance=False)

    # Min_samples less than one
    with pytest.raises(ValueError,
                       match="min_samples may not be larger than max_neighbors_to_calculate"):
        ocelot.cluster.epsilon.acg18(
            data_rescaled, distance_matrix, n_repeats=2, min_samples=0, return_last_random_distance=False)

    return epsilons, random_distance_matrix


def test__summed_kth_nn_distribution_one_cluster():
    """Tests ocelot.cluster.epsilon_summed_kth_nn_distribution_one_cluster (and by extension, kth_nn_distribution)."""
    # Set some parameters to play with
    field_constant = 0.3
    field_dimension = 5
    cluster_constant = 0.05
    cluster_dimension = 3
    cluster_fraction = 0.01
    parameters = np.asarray([field_constant, field_dimension, cluster_constant, cluster_dimension, cluster_fraction])

    # Calculate some residuals
    residual_inf = ocelot.cluster.epsilon._summed_kth_nn_distribution_one_cluster(
        parameters, 10, np.linspace(0.0, 1, num=50), y_range=np.linspace(1, 1000, num=50) ** 2, minimisation_mode=True)

    residual_good = ocelot.cluster.epsilon._summed_kth_nn_distribution_one_cluster(
        parameters, 10, np.linspace(0.1, 1, num=50), y_range=np.linspace(1, 1000, num=50) ** 2, minimisation_mode=True)

    # Check that the bad residual is inf
    assert residual_inf == np.inf

    # Check that the good residual is... good
    assert np.allclose(residual_good, 10316599967493.814, rtol=0.001, atol=1e-8)

    # Calculate a field
    y_fields = ocelot.cluster.epsilon._summed_kth_nn_distribution_one_cluster(
        parameters, 10, np.linspace(0, 1, num=50), minimisation_mode=False)

    # Check the shape
    assert y_fields.shape == (3, 50)

    # Check that we get np.inf on the first value, showing that my log comprehension works
    assert np.allclose(y_fields[:, 0], np.asarray([-np.inf, -np.inf, -np.inf]), rtol=0.0, atol=1e-8)

    # Check that the mean of everything else is right, and by extension it's... probably right
    assert np.allclose(np.mean(y_fields[:, 1:]), -1.4524090935710814, rtol=0.0, atol=1e-8)

    return 0


def test__find_sign_change_epsilons():
    """Tests ocelot.cluster.epsilon._find_sign_change_epsilons, a function that finds the biggest second derivative
    bump.

    Diagnostic code to plot what the function sees:

    import matplotlib.pyplot as plt
    x = np.log10(x)
    plt.plot(x, np.gradient(np.gradient(y, x), x), 'r-')
    plt.show()

    print(epsilon_values, all_sign_changes)

    """

    # Make some fake data - a squared sine wave that gets exponentially smaller
    x = np.geomspace(0.1, 40 * np.pi, num=200)
    y = (np.sin(np.log10(x)) * np.exp(-x / 10)) ** 2 - 0.2

    # Call the function and hope for a good result lol
    epsilon_values, all_sign_changes = ocelot.cluster.epsilon._find_sign_change_epsilons(
        x, y, return_all_sign_changes=True)

    # Check that it got the right number of sign changes
    assert all_sign_changes.shape == (3,)

    # Check that the main sign changes are right (could be sensitive to setting changes in the resolution of x)
    target = [0.15938995421984642, 0.6934124282589442, 2.2642842494315762]
    assert np.allclose(epsilon_values, target, rtol=0.0, atol=1e-8)


def test__find_curve_absolute_maximum_epsilons():
    """Tests ocelot.cluster.epsilon._find_curve_absolute_maximum_epsilons

    Test plotting code that shows what the function sees::

    import matplotlib.pyplot as plt
    x = np.log10(x)
    plt.plot(x, np.abs(np.gradient(np.gradient(y, x, ), x, )))
    plt.show()

    print(maximum_x)

    """
    # Make some fake data
    x = np.linspace(0, 10, num=200)
    y = x**2 * np.exp(-x) - 1

    # Call time
    maximum_x = ocelot.cluster.epsilon._find_curve_absolute_maximum_epsilons(x, y)

    assert np.allclose(maximum_x, 2.361809045226131, rtol=0.0, atol=1e-8)


def test_field_model(show_figure=False):
    """Tests ocelot.cluster.epsilon.field_model using data of Blanco 1. It can be worth testing this manually too,
    as it does quite a lot of funky stuff that's worth looking at."""
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Re-scale the data with standard scaling and check that it has zero mean, unit variance
    data_rescaled = ocelot.cluster.rescale_dataset(data_gaia, scaling_type='robust')

    # Calculate some nearest neighbor distances baby!
    distance_matrix = ocelot.cluster.precalculate_nn_distances(
        data_rescaled, n_neighbors=10, return_sparse_matrix=False, return_knn_distance_array=True)

    # Specify some plot options
    plot_options = {'number_of_derivatives': 2,
                    'figure_size': (6, 8),
                    'show_figure': show_figure,
                    'figure_title': 'Unit test of ocelot.cluster.epsilon.field_model',
                    }

    # See what the field model fit thinks of this
    success, epsilon_values, parameters, n_cluster_members = ocelot.cluster.epsilon.field_model(
        distance_matrix, min_samples=10, min_cluster_size=1, make_diagnostic_plot=True,
        **plot_options)

    # Test that the results are good
    assert success is True

    target_epsilon = [0.13765954, 0.1809968, 0.22433406, 0.25747432, 0.28968726]
    assert np.allclose(epsilon_values, target_epsilon, rtol=0.0, atol=1e-8)

    target_parameters = [0.26971331, 9.40067326, 0.08953837, 3.36263488, 0.02072574]
    assert np.allclose(parameters, target_parameters, rtol=0.0, atol=1e-8)

    assert n_cluster_members == 291


def test_read_cmd_simulated_populations():
    """Tests the ability of ocelot.cluster.SimulatedPopulations to read in CMD simulated populations in bulk."""
    target_columns = ['Z', 'age', 'Mini', 'Mass', 'logL', 'logTe', 'logg', 'cmd_label', 'mbolmag',
                      'phot_g_mean_mag', 'phot_bp_mean_mag', 'phot_rp_mean_mag', 'log_age', 'log_Z']

    # Check that we can read in one population ok
    data_one = ocelot.cluster.SimulatedPopulations(path_to_one_simulated_population).data

    assert data_one.shape == (6116, 14)
    assert np.allclose(data_one.iloc[1500, 5], 3.7996)
    assert list(data_one.columns) == target_columns

    # Check that we can read in multiple
    data_all = ocelot.cluster.SimulatedPopulations(path_to_all_simulated_populations, search_pattern="*.dat").data

    assert data_all.shape == (2 * 6116, 14)
    assert np.allclose(data_all.iloc[6116 + 1500, 5], 3.7996)
    assert list(data_all.columns) == target_columns

    return data_one, data_all


def test_simulated_populations(plot_clusters=False):
    """Given that ocelot.cluster.SimulatedPopulations reading works (see test_read_cmd_simulated_populations),
    we next test whether or not it can make some simulated populations!

    Data generated with:
        data_smol = data_mwsc.loc[clusters_to_plot_mwsc_table_numbers, :]
        data_smol['extinction_v'] = 3.1 * data_smol['E(B-V)']
        data_smol['v_int'] = 500
        data_smol['[Fe/H]'] = 0.
        data_smol['mass'] = 1e3
        new_data = pd.DataFrame()
        new_data[['ra', 'dec', 'l', 'b', 'distance', 'extinction_v', 'pmra', 'pmdec',
                  'age', 'Z', 'mass', 'radius_c', 'radius_t', 'v_int']] = data_smol[[
            'RAJ2000', 'DEJ2000', 'GLON', 'GLAT', 'd', 'extinction_v', 'pmRA', 'pmDE',
            'logt', '[Fe/H]', 'mass', 'rc', 'rt', 'v_int'
        ]]
        new_data.to_csv('test_clusters', index=False)

    Todo: expand this unit test to be more than just a visual one

    """
    np.random.seed(42)
    test_clusters = pd.read_csv(path_to_simulated_population_test_clusters)

    simulated_populations = ocelot.cluster.SimulatedPopulations(path_to_big_simulated_population,
                                                                mass_tolerance=0.01)

    simulated_clusters = simulated_populations.get_clusters(test_clusters, concatenate=False,
                                                            error_on_invalid_request=False,)

    # Plot some clustery friends for the user, if requested
    if plot_clusters:

        # Plot all the clusters
        for number, a_cluster in enumerate(simulated_clusters):
            ocelot.plot.clustering_result(a_cluster, plot_std_limit=4., figure_title=f"test cluster {number}")

        # Grab Blanco 1 and correct for its RAs (which straddle 0)
        test_cluster = simulated_clusters[0]
        with open(Path('./test_data/blanco_1_gaia_dr2_gmag_18_cut_cluster_only.pickle'), 'rb') as handle:
            data_gaia = pickle.load(handle)
        data_gaia['ra'] = np.where(data_gaia['ra'] > 180, data_gaia['ra'] - 360, data_gaia['ra'])

        # Plotting time of the King profile - Blanco 1 vs a simulated version!
        plt.hist(np.sqrt((test_cluster['ra'] - np.mean(test_cluster['ra'])) ** 2
                         + (test_cluster['dec'] - np.mean(test_cluster['dec'])) ** 2),
                 bins='auto', density=True, color='k', alpha=0.5, label='generated', cumulative=False)
        plt.hist(np.sqrt(
            (data_gaia['ra'] - np.mean(data_gaia['ra'])) ** 2 + (data_gaia['dec'] - np.mean(data_gaia['dec'])) ** 2),
                 bins='auto', density=True, color='b', alpha=0.5, label='blanco 1', cumulative=False)
        plt.yscale('log')
        plt.legend()
        plt.xlabel('r')
        plt.ylabel('log normalised density')
        plt.show()

    return simulated_populations, test_clusters, simulated_clusters


def test__setup_cluster_parameter():
    """Tests ocelot.cluster.synthetic._setup_cluster_parameter, a function for handling cluster parameter stuff."""
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Constants
    parameter_name = 'ra'

    # Test identical parameters, for both int and float specified - both cases should return an array of floats
    target = np.ones(5, dtype=float) * 2.

    float_like = ocelot.cluster.synthetic._setup_cluster_parameter(data_gaia, 'ra', 2., 5)
    assert np.allclose(target, float_like)

    int_like = ocelot.cluster.synthetic._setup_cluster_parameter(data_gaia, 'ra', 2, 5)
    assert np.allclose(target, int_like)

    # Test sampling of parameters
    np.random.seed(42)
    target = np.asarray([0.75, 1., 0.5])

    draw_from = np.linspace(0, 1, num=5)
    array_like = ocelot.cluster.synthetic._setup_cluster_parameter(data_gaia, 'ra', draw_from, 3)

    assert np.allclose(target, array_like)

    # Test callable function use
    target = np.asarray([-3.98379464, 5.76513088])

    callable_like = ocelot.cluster.synthetic._setup_cluster_parameter(
        data_gaia, 'ra', ocelot.cluster.synthetic._c_position_limits_plus_minus_two, 2)

    assert np.allclose(target, callable_like)


def test__find_nearest_magnitude_star():
    """Tests ocelot.cluster.synthetic._find_nearest_magnitude_star, which basically just does some array maths"""
    real = np.linspace(0, 10, num=1000)
    synthetic = np.linspace(0, 10, num=6)
    np.random.seed(42)

    matches = ocelot.cluster.synthetic._find_nearest_magnitude_star(real, synthetic)

    # Check that we get the correct number of matches
    assert synthetic.shape == matches.shape

    # Check that the values returned are about the same
    assert np.allclose(synthetic, real[matches], atol=0.01)


def test_generate_synthetic_clusters(plot_clusters=True):
    """Tests ocelot.cluster.generate_synthetic_clusters, a fancy wrapper to the SimulatedPopulations class that also
    handles adding things like error."""
    # Read in data for Blanco 1 and make a simulated populations object
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Add some random errors onto the data because Emily was a doofus and didn't download Gaia data with flux errors
    data_gaia['phot_g_mean_flux_error'] = 4.
    data_gaia['phot_bp_mean_flux_error'] = 30.
    data_gaia['phot_rp_mean_flux_error'] = 30.

    simulated_populations = ocelot.cluster.SimulatedPopulations(path_to_big_simulated_population,
                                                                mass_tolerance=0.01)

    # Make some clusters in augmentation input_mode
    data_augmented = ocelot.cluster.generate_synthetic_clusters(
        simulated_populations,
        data_gaia,
        mode='clustering_augmentation',
        cluster_parameters_to_overwrite={'age': 9.5},
        shuffle=False,)

    data_augmented = ocelot.cluster.cut_dataset(data_augmented, parameter_cuts={'phot_g_mean_mag': [-np.inf, 18]})
    data_augmented['ra'] = np.where(data_augmented['ra'] > 180, data_augmented['ra'] - 360, data_augmented['ra'])

    # Make some clusters in generator input_mode
    import time
    start = time.time()
    data_generated = ocelot.cluster.generate_synthetic_clusters(
        simulated_populations,
        data_gaia,
        mode='generator',
        cluster_parameters_to_overwrite={'distance': ocelot.cluster.synthetic._c_random_cbj_distance,
                                         'age': np.linspace(6., 9.5, num=50),
                                         'v_int': np.random.normal(loc=500, scale=50, size=50),
                                         'mass': np.random.exponential(scale=200, size=50) + 50,
                                         'extinction_v': np.random.exponential(scale=1., size=20)},
        shuffle=True,
        concatenate=False,
        n_clusters=10)
    print(time.time() - start)

    data_generated = ocelot.cluster.cut_dataset(data_generated, parameter_cuts={'phot_g_mean_mag': [-np.inf, 21]})
    data_generated['ra'] = np.where(data_generated['ra'] > 180, data_generated['ra'] - 360, data_generated['ra'])

    if plot_clusters:
        ocelot.plot.clustering_result(data_augmented,
                                      data_augmented['cluster_label'],
                                      [0, 1],
                                      plot_std_limit=5.,
                                      figure_title='augmented clustering',
                                      cmd_plot_y_limits=[8, 18])

        ocelot.plot.clustering_result(data_generated,
                                      data_generated['cluster_label'],
                                      np.unique(data_generated['cluster_label']),
                                      plot_std_limit=1.5,
                                      figure_title='generated clustering',
                                      cmd_plot_y_limits=[8, 21])


if __name__ == '__main__':
    # pd.set_option('display.max_columns', None)
    # pd.set_option('display.max_rows', None)
    # cut = test_cut_dataset()
    # gaia, rescaled = test_rescale_dataset()
    # spar, dist = test_precalculate_nn_distances()
    # eps, ran = test_acg18()
    # test__summed_kth_nn_distribution_one_cluster()
    # test__find_sign_change_epsilons()
    # test__find_curve_absolute_maximum_epsilons()
    # test_field_model(show_figure=True)
    # one, all = test_read_cmd_simulated_populations()
    # simpop, test_clusters, simcl = test_simulated_populations(plot_clusters=False)
    # test__setup_cluster_parameter()
    # test__find_nearest_magnitude_star()
    # test_generate_synthetic_clusters(plot_clusters=True)
    # gaia = test_recenter_dataset(show_figure=True)
    # gaia = test_recenter_dataset_healpix(show_figure=True).
    # largest = test_find_largest_cluster()
    data, parter = test_data_partition(show_figure=True)



"""Utilities for checking that things have the correct shape, etc."""


def _check_matching_lengths_of_non_nones(*args):
    """Checks that all non-None arguments have the same length."""
    if len(args) <= 1:
        return

    args_excluding_nones = [arg for arg in args if arg is not None]
    first_length = len(args_excluding_nones[0])
    for i, length in enumerate(args_excluding_nones[1:]):
        if length != first_length:
            raise ValueError(
                f"Length of argument {i+1} does not match length of first element"
            )



"""A set of functions for calculating typical cluster parameters.

Todo: error treatment here could be made more bayesian
"""

from typing import Optional
from astropy.coordinates import SkyCoord

import numpy as np
import pandas as pd

from .constants import mas_per_yr_to_rad_per_s, pc_to_m, deg_to_rad


def _handle_ra_discontinuity(ra_data, middle_ras_raise_error=True):
    """Tries to detect when the ras in a field cross the (0, 360) ra discontinuity and returns corrected results. Will
    raise an error if ras are all over the place (which will happen e.g. at very high declinations) in which you
    ought to instead switch to a method free of spherical distortions.

    Args:
        ra_data (pd.Series or np.ndarray): data on ras.
        middle_ras_raise_error (bool): whether or not a cluster having right ascensions in all ranges [0, 90), [90, 270]
            and (270, 360] raises an error. The error here indicates that this cluster has extreme spherical
            discontinuities (e.g. it's near a coordinate pole) and that the mean ra and mean dec will be inaccurate.
            Default: True

    Returns:
        ra_data but corrected for distortions. If values are both <90 and >270, the new ra data will be in the range
            (-90, 90).

    """
    # Firstly, check that the ras are valid ras
    if np.any(ra_data >= 360) or np.any(ra_data < 0):
        raise ValueError(
            "at least one input ra value was invalid! Ras must be in the range [0, 360)."
        )

    # Next, grab all the locations of dodgy friends and check that all three aren't ever in use at the same time
    low_ra = ra_data < 90
    high_ra = ra_data > 270
    middle_ra = np.logical_not(np.logical_or(low_ra, high_ra))

    # Proceed if we have both low and high ras
    if np.any(low_ra) and np.any(high_ra):
        # Stop if we have middle too (would imply stars everywhere or an extreme dec value)
        if np.any(middle_ra) and middle_ras_raise_error:
            raise ValueError(
                "ra values are in all three ranges: [0, 90), [90, 270] and (270, 360). This cluster can't "
                "be processed by this function! Spherical distortions must be removed first."
            )

        # Otherwise, apply the discontinuity removal
        else:
            # Make a copy so nothing weird happens
            ra_data = ra_data.copy()

            # And remove the distortion for all high numbers
            ra_data[high_ra] = ra_data[high_ra] - 360

    return ra_data


def mean_radius(
    data_gaia: pd.DataFrame,
    membership_probabilities: Optional[np.ndarray] = None,
    already_inferred_parameters: Optional[dict] = None,
    key_ra: str = "ra",
    key_ra_error: str = "ra_error",
    key_dec: str = "dec",
    key_dec_error: str = "dec_error",
    distance_to_use: str = "inverse_parallax",
    middle_ras_raise_error: bool = True,
    **kwargs,
):
    """Produces various radius statistics on a given cluster, finding its sky location and three radii: the core, tidal
    and 50% radius.

    Done in a very basic, frequentist way, whereby means are weighted based on the membership probabilities (if
    specified).

    N.B. unlike the above functions, errors do *not change the mean* as this would potentially bias the
    estimates towards being dominated by large, centrally-located stars within clusters (that have generally lower
    velocities.) Hence, estimates here will be less precise but hopefully more accurate.

    Todo: add error estimation to this function (hard)

    Todo: add galactic l, b to the output of this function

    Args:
        data_gaia (pd.DataFrame): Gaia data for the cluster in the standard format (e.g. as in DR2.)
        membership_probabilities (optional, np.ndarray): membership probabilities for the cluster. When specified,
            they can increase or decrease the effect of certain stars on the mean.
        already_inferred_parameters (optional, dict): a parameter dictionary of the mean distance and proper motion.
            Otherwise, this function calculates a version.
        key_ra (str): Gaia parameter name.
        key_ra_error (str): Gaia parameter name.
        key_dec (str): Gaia parameter name.
        key_dec_error (str): Gaia parameter name.
        distance_to_use (str): which already inferred distance to use to convert angular radii to parsecs.
            Default: "inverse_parallax"
        middle_ras_raise_error (bool): whether or not a cluster having right ascensions in all ranges [0, 90), [90, 270]
            and (270, 360] raises an error. The error here indicates that this cluster has extreme spherical
            discontinuities (e.g. it's near a coordinate pole) and that the mean ra and mean dec will be inaccurate.
            Default: True

    Returns:
        a dict, formatted with:
            {
            # Position
            "ra": ra of the cluster
            "ra_error": error on the above
            "dec": dec of the cluster
            "dec_error": error on the above

            # Angular radii
            "ang_radius_50": median ang. distance from the center, i.e.angular radius of the cluster with 50% of members
            "ang_radius_50_error": error on the above
            "ang_radius_c": angular King's core radius of the cluster
            "ang_radius_c_error": error on the above
            "ang_radius_t": maximum angular distance from the center, i.e. angular King's tidal radius of the cluster
            "ang_radius_t_error": error on the above

            # Parsec radii
            "radius_50": median distance from the center, i.e.radius of the cluster with 50% of members
            "radius_50_error": error on the above
            "radius_c": King's core radius of the cluster
            "radius_c_error": error on the above
            "radius_t": maximum distance from the center, i.e. King's tidal radius of the cluster
            "radius_t_error": error on the above
            }

    """
    inferred_parameters = {}
    sqrt_n_stars = np.sqrt(data_gaia.shape[0])

    # Grab the distances if they aren't specified - we'll need them in a moment!
    if already_inferred_parameters is None:
        already_inferred_parameters = mean_distance(data_gaia, membership_probabilities)

    # Estimate the ra, dec of the cluster as the weighted mean
    ra_data = _handle_ra_discontinuity(
        data_gaia[key_ra], middle_ras_raise_error=middle_ras_raise_error
    )

    inferred_parameters["ra"] = np.average(ra_data, weights=membership_probabilities)
    inferred_parameters["ra_std"] = _weighted_standard_deviation(
        ra_data, membership_probabilities
    )
    inferred_parameters["ra_error"] = inferred_parameters["ra_std"] / sqrt_n_stars

    if inferred_parameters["ra"] < 0:
        inferred_parameters["ra"] += 360

    inferred_parameters["dec"] = np.average(
        data_gaia[key_dec], weights=membership_probabilities
    )
    inferred_parameters["dec_std"] = _weighted_standard_deviation(
        data_gaia[key_dec], membership_probabilities
    )
    inferred_parameters["dec_error"] = inferred_parameters["dec_std"] / sqrt_n_stars

    inferred_parameters["ang_dispersion"] = np.sqrt(
        inferred_parameters["ra_std"] ** 2 + inferred_parameters["dec_std"] ** 2
    )

    # Calculate how far every star in the cluster is from the central point
    center_skycoord = SkyCoord(
        ra=inferred_parameters["ra"], dec=inferred_parameters["dec"], unit="deg"
    )
    star_skycoords = SkyCoord(
        ra=data_gaia[key_ra].to_numpy(), dec=data_gaia[key_dec].to_numpy(), unit="deg"
    )

    distances_from_center = center_skycoord.separation(star_skycoords).degree

    # And say something about the radii in this case
    inferred_parameters["ang_radius_50"] = np.median(distances_from_center)
    inferred_parameters["ang_radius_50_error"] = np.nan

    inferred_parameters["ang_radius_c"] = np.nan
    inferred_parameters["ang_radius_c_error"] = np.nan

    inferred_parameters["ang_radius_t"] = np.max(distances_from_center)
    inferred_parameters["ang_radius_t_error"] = np.nan

    # Convert the angular distances into parsecs
    inferred_parameters["radius_50"] = (
        np.tan(inferred_parameters["ang_radius_50"] * deg_to_rad)
        * already_inferred_parameters[distance_to_use]
    )
    inferred_parameters["radius_50_error"] = np.nan
    inferred_parameters["radius_c"] = np.nan
    inferred_parameters["radius_c_error"] = np.nan
    inferred_parameters["radius_t"] = (
        np.tan(inferred_parameters["ang_radius_t"] * deg_to_rad)
        * already_inferred_parameters[distance_to_use]
    )
    inferred_parameters["radius_t_error"] = np.nan

    return inferred_parameters


def all_statistics():
    """
    """
    # Todo refactor this (and other high-level calculation methods)
    pass



"""Various possible axes for use in the analysis of nearest neighbor distances (e.g. when running DBSCAN.)"""

from typing import Optional, Union

import numpy as np

from ..utilities import normalise_a_curve, percentile_based_plot_limits


def point_number_vs_nn_distance(axes, distances: np.ndarray,
                                neighbor_to_plot: int = 1,
                                normalisation_constants: Union[list, tuple, np.ndarray] = (1.,),
                                functions_to_overplot: Optional[list] = None,
                                show_numerical_derivatives: bool = False,
                                y_percentile_limits: Optional[np.ndarray] = None, ):
    """Plots point number vs nearest neighbor distance, with the option to add multiple regions to the plot (like
    highlighting an area of maximum curvature.)

    Notes:
        - to overplot a fitting function, all of fitting_functions_x, fitting_functions_y and fitting_function_styles
            must be specified.

    Args:
        axes (list of matplotlib axes): the axes to plot on. Each additional axis will have another level of
            derivative plotted on it.
        distances (np.ndarray): the kth nearest neighbor distance array, of shape (n_samples, n_neighbors_calculated.)
        neighbor_to_plot (int): kth nearest neighbor to plot. Must be lower than n_neighbors_calculated.
            Default: 1
        normalisation_constants (list-like of floats): normalisation constants to use for the curves. If zero,
            no normalisation will be performed. If a single float, all curves will be normalised to the same area. If a
            list of floats, all curves will have a different normalisation constant applied (extremely useful for
            plotting multiple components of a single model curve.) Normalisation is performed by the trapezium rule.
            Default: 1., i.e. all curves normalised to area 1
        functions_to_overplot (optional, list of dictionaries): list of fitting function information to overplot. Each
            dictionary should have:
            'x': x values (not logged)
            'y': y values (not logged)
            'style': matplotlib style, e.g. 'r--'
            'label': label for the curve or None
            'differentiate': whether to differentiate this each time (True/False). You probably want True, unless this
                is a specific value.
            Default: None
        show_numerical_derivatives (bool): whether or not to show the numerical derivatives of the field curve. They are
            typically very noisy so this isn't desired behaviour.
            Default: False
        y_percentile_limits (optional, list-like): percentile limits to apply to each derivative axis. Should be a
            np.ndarray of shape (n_extra_features - 1, 2), containing floats in the range [0, 100] and strictly ascending. Will be
            ignored for all axes that contain real points.

    Returns:
         the modified matplotlib axis.

    """
    # Is the user a fuckwit? Let's find out!
    if neighbor_to_plot > distances.shape[1]:
        raise ValueError("neighbor_to_plot cannot be higher than the number of calculated nearest neighbors!")

    # Count the number of fitting functions
    if functions_to_overplot is not None:
        n_fitting_functions = len(functions_to_overplot)
    else:
        n_fitting_functions = 0

    # Make sure that y_percentile_limits exists if it needs to if the user hasn't already specified limits
    if show_numerical_derivatives is False and n_fitting_functions > 0:
        infer_percentile_limits = True
        if y_percentile_limits is None:
            y_percentile_limits = np.tile(np.asarray([0, 100]), (n_fitting_functions, 1))
    else:
        infer_percentile_limits = False

    # Make sure the axes are a list and count the number of axes
    if type(axes) is not np.ndarray:
        axes = np.asarray([axes])
    n_axes = len(axes)

    # Grab some useful values
    nearest_neighbor_distances = np.sort(distances[:, neighbor_to_plot - 1])
    point_numbers = np.arange(nearest_neighbor_distances.shape[0]) + 1

    # Normalisation & log time
    point_numbers = normalise_a_curve(nearest_neighbor_distances, point_numbers, normalisation_constants[0])
    nearest_neighbor_distances = np.log10(nearest_neighbor_distances)
    point_numbers = np.log10(point_numbers)

    # Normalise and log the fitting functions too! We do this safely as there are more ways in which this could be
    # passed an invalid value expected (instead of the actually-physical real distances).
    i = 0
    while i < n_fitting_functions:
        # We'll only want to work on good stars
        good_stars = np.logical_and(
            np.logical_and(np.isfinite(functions_to_overplot[i]['x']), np.asarray(functions_to_overplot[i]['x']) > 0),
            np.logical_and(np.isfinite(functions_to_overplot[i]['y']), np.asarray(functions_to_overplot[i]['y']) > 0))

        functions_to_overplot[i]['x'] = np.asarray(functions_to_overplot[i]['x'])[good_stars]
        functions_to_overplot[i]['y'] = np.asarray(functions_to_overplot[i]['y'])[good_stars]

        # Normalise the y values
        functions_to_overplot[i]['y'] = np.asarray(normalise_a_curve(
            functions_to_overplot[i]['x'], functions_to_overplot[i]['y'], normalisation_constants[i + 1]))
        functions_to_overplot[i]['x'] = np.asarray(functions_to_overplot[i]['x'])

        # We'll only want to work on good stars (checking again just to be super safe lol)
        good_y_values = functions_to_overplot[i]['y'] > 0
        good_x_values = functions_to_overplot[i]['x'] > 0

        # Take logs only where log() is defined, otherwise replace with -np.inf
        functions_to_overplot[i]['y'] = np.where(
            good_y_values, np.log10(functions_to_overplot[i]['y'], where=good_y_values), -np.inf)

        functions_to_overplot[i]['x'] = np.where(
            good_x_values, np.log10(functions_to_overplot[i]['x'], where=good_x_values), -np.inf)

        i += 1

    # Cycle over axes, making my way downtown (to the-plot-is-done-land)
    x_limits = None  # Just here to shut the fucking LINTER UP
    for i_derivative, an_axis in enumerate(axes):
        # We keep hold of all the x and y data in a single array each round so that the limits can be correctly
        # calculated in a moment
        all_x_data = []
        all_y_data = []

        # We only need to take derivatives if we aren't on the final axis
        do_we_need_to_take_derivatives = i_derivative < n_axes - 1

        # Plot the field stars
        if i_derivative == 0 or show_numerical_derivatives:
            axes[i_derivative].plot(nearest_neighbor_distances, point_numbers, 'k-', label='Field stars')

            # Take derivatives while we're here, if necessary
            if do_we_need_to_take_derivatives and show_numerical_derivatives:
                point_numbers = np.gradient(point_numbers, nearest_neighbor_distances)

        # Plot the fitting functions
        i = 0
        while i < n_fitting_functions:
            axes[i_derivative].plot(functions_to_overplot[i]['x'], functions_to_overplot[i]['y'],
                                    functions_to_overplot[i]['style'], label=functions_to_overplot[i]['label'])

            # Only base the limits off of the stuff we're diffr'ing
            if functions_to_overplot[i]['differentiate']:
                all_x_data.append(functions_to_overplot[i]['x'].flatten())
                all_y_data.append(functions_to_overplot[i]['y'].flatten())

                # Take derivatives while we're here, if necessary (aka if this isn't the last axis) and remove any
                # gradient values that will have become inf
                if do_we_need_to_take_derivatives:
                    good_stars = np.logical_and(np.isfinite(functions_to_overplot[i]['x']),
                                                np.isfinite(functions_to_overplot[i]['y']))
                    functions_to_overplot[i]['x'] = (functions_to_overplot[i]['x'])[good_stars]
                    functions_to_overplot[i]['y'] = (functions_to_overplot[i]['y'])[good_stars]

                    functions_to_overplot[i]['y'] = np.gradient(functions_to_overplot[i]['y'],
                                                                functions_to_overplot[i]['x'])
            i += 1

        # Set y and x limits based on the points if they're in the plot (always the case in the first one,) or based on
        # the fitting functions if not
        if i_derivative == 0 or i == 0:
            x_limits = [np.min(nearest_neighbor_distances)-0.1, np.max(nearest_neighbor_distances)+0.1]
            y_limits = [np.min(point_numbers) - 1, np.max(point_numbers) + 1]
        elif infer_percentile_limits is False:
            y_limits = [np.min(point_numbers) - 1, np.max(point_numbers) + 1]
        else:
            all_x_data = np.concatenate(all_x_data)
            all_y_data = np.concatenate(all_y_data)
            x_limits_from_percentiles, y_limits = percentile_based_plot_limits(
                all_x_data, all_y_data, y_percentiles=y_percentile_limits[i_derivative - 1])

        axes[i_derivative].set_xlim(x_limits)
        axes[i_derivative].set_ylim(y_limits)

        # Axis beautification
        axes[i_derivative].minorticks_on()  # Makes it easier to read!
        axes[i_derivative].set_xlabel(f"log {neighbor_to_plot}th nearest neighbour distance")

        if i_derivative == 0:
            axes[i_derivative].set_ylabel(f"log point number")
        else:
            axes[i_derivative].set_ylabel(f"log point number\n{i_derivative}th derivative")

    axes[0].legend(fontsize=8, fancybox=True, edgecolor='black', loc='lower right')
    return axes



"""Different methods for calculating the center of a cluster in spherical coordinates. 
(This is actually oddly difficult, thanks to how spheres work. Damn spheres.)
"""
import numpy as np
from numpy.typing import ArrayLike

from ocelot.calculate.generic import spherical_mean


def mean_position(
    longitudes: ArrayLike, latitudes: ArrayLike, degrees=True
) -> tuple[float]:
    """Calculates the spherical mean of angular positions, specified as longitudes and
    latitudes. This uses directional statistics to do so in a way that is aware of
    discontinuities, such as the fact that 0° = 360°.

    Parameters
    ----------
    longitudes : array-like
        Array of longitudinal positions of stars in your cluster (e.g. right ascensions
        or galactic longitudes.) Assumed to be in the range [0°, 360°].
    latitudes : array-like
        Array of latitudinal positions of stars in your cluster (e.g. declinations or 
        galactic latitudes.) Assumed to be in the range [-90°, 90°].
    degrees : bool
        Whether longitudes and latitudes are in degrees, and whether to return an answer
        in degrees. Defaults to True. If False, longitudes and latitudes are assumed to
        be in radians, with ranges [0, 2π] and [-π/2, π/2] respectively.

    Returns
    -------
    mean_longitude : float
    mean_latitude : float

    Notes
    -----
    This function explicitly assumes that your star cluster *has* a well-defined mean
    position. Some configurations (such as points uniformly distributed in at least one
    axis of a sphere) will not have a meaningful mean position.

    Internally, this function uses `scipy.stats.directional_stats`, with a definition
    taken from [1]. See [2] for more background.

    References
    ----------
    [1] Mardia, Jupp. (2000). Directional Statistics (p. 163). Wiley.
    [2] https://en.wikipedia.org/wiki/Directional_statistics
    """
    if degrees:
        longitudes, latitudes = np.radians(longitudes), np.radians(latitudes)
    mean_lon, mean_lat = spherical_mean(longitudes, latitudes)
    if degrees:
        return np.degrees(mean_lon), np.degrees(mean_lat)
    return mean_lon, mean_lat


def mode_position():
    """Attempts to find the mode of a star cluster's 2D on-sky distribution. This is a 
    better estimator than the mean position for clusters that are assymmetric, which is
    often the case for clusters with assymetric tidal tails (e.g. due to one side being
    more easily detected than the other.)
    """
    # Todo
    pass
